from typing import Callable

import numpy as np

from sampo.scheduler.resource.base import ResourceOptimizer
from sampo.scheduler.utils import WorkerContractorPool
from sampo.schemas.resources import Worker
from sampo.schemas.time import Time
from sampo.utilities.base_opt import coordinate_descent


class CoordinateDescentResourceOptimizer(ResourceOptimizer):
    """
    Class that implements optimization the number of resources by discrete analogue of coordinate descent.
    """

    def __init__(self, one_dimension_optimizer: Callable[[int, int, Callable[[int], Time]], Time]):
        self.one_dimension_optimizer = one_dimension_optimizer

    def optimize_resources(self,
                           worker_pool: WorkerContractorPool,
                           worker_team: list[Worker],
                           optimize_array: np.ndarray,
                           down_border: np.ndarray,
                           up_border: np.ndarray,
                           get_finish_time: Callable[[list[Worker]], Time]):
        """
        The resource optimization module, that search optimal number of resources by coordinate descent.

        :param worker_pool: global resources pool
        :param worker_team: worker team to optimize
        :param optimize_array: a boolean array that says what positions should be optimized
        :param down_border: down border of optimization
        :param up_border: up border of optimization
        :param get_finish_time: optimization function that should give execution time based on the worker team
        """

        def fitness(worker_count: np.ndarray):
            """
            Function that counts finish time with a current set of resources.

            :param worker_count:
            :return:
            """
            for ind, worker in enumerate(worker_team):
                worker.count = worker_count[ind]
            return get_finish_time(worker_team)

        count_worker_team = coordinate_descent(down_border, up_border,
                                               self.one_dimension_optimizer,
                                               fitness,
                                               optimize_array)
        if optimize_array:
            for i, worker in enumerate(worker_team):
                if optimize_array[i]:
                    worker.count = count_worker_team[i]
        else:
            for i, worker in enumerate(worker_team):
                worker.count = count_worker_team[i]



from sampo.scheduler.resource.average_req import AverageReqResourceOptimizer
from sampo.scheduler.resource.coordinate_descent import CoordinateDescentResourceOptimizer
from sampo.scheduler.resource.full_scan import FullScanResourceOptimizer
from sampo.scheduler.resource.identity import IdentityResourceOptimizer



from copy import deepcopy
from typing import Iterable

from sampo.scheduler.multi_agency.block_graph import BlockGraph
from sampo.scheduler.multi_agency.multi_agency import ScheduledBlock, Agent
from sampo.schemas.contractor import Contractor
from sampo.utilities.validation import validate_schedule, \
    check_all_allocated_workers_do_not_exceed_capacity_of_contractors


def validate_block_schedule(bg: BlockGraph, schedule: dict[str, ScheduledBlock], agents: Iterable[Agent]):
    contractors = [contractor for agent in agents for contractor in agent.contractors]
    contractors_set = set(contractors)

    assert len(contractors) == len(contractors_set), \
        f'There are contractor collisions between agents: ' \
        f'{[c.id for c in contractors]} != {[c.id for c in contractors_set]}'

    _check_block_dependencies(bg, schedule)
    _check_blocks_separately(schedule.values())

    # TODO Fix. To fully validate resources usage in the whole multi-agent appearance,
    #    we should union all agent's ScheduleEvents and go through it in sorted way.
    #    Now we think that this validation phase is not extremely need.
    # _check_blocks_with_global_timelines(schedule.values(), contractors)


def _check_block_dependencies(bg: BlockGraph, schedule: dict[str, ScheduledBlock]):
    """
    Validate right block dependencies considering received 'schedule'

    :param bg: BlockGraph
    :param schedule:
    """
    for sblock in schedule.values():
        for p in bg[sblock.id].blocks_from:
            parent_sblock = schedule[p.id]
            assert parent_sblock.start_time <= parent_sblock.end_time <= sblock.start_time <= sblock.end_time


def _check_blocks_separately(sblocks: Iterable[ScheduledBlock]):
    """
    Validate each block separately, i.e., check each 'Schedule' in each block

    :param sblocks: scheduled blocks of works
    """
    for sblock in sblocks:
        try:
            validate_schedule(sblock.schedule, sblock.wg, sblock.agent.contractors)
        except AssertionError as e:
            raise AssertionError(f'Agent {sblock.agent} supplied an invalid schedule') from e


def _check_blocks_with_global_timelines(sblocks: Iterable[ScheduledBlock], contractors: Iterable[Contractor]):
    """
    Checks that no agent's contractor uses more resources that can supply.

    Note that this should fail if there is a shared contractor between agents, but this
    term is, of course, unsupported.

    :param sblocks: scheduled blocks of works
    :param contractors: global scope of contractors(collected from all agents used to construct sblocks)
    """
    initial_worker_pool: dict[str, dict[str, int]] = {}
    for contractor in contractors:
        initial_worker_pool[contractor.id] = {}
        for w in contractor.workers.values():
            if w.name in initial_worker_pool[contractor.id].keys():
                initial_worker_pool[contractor.id][w.name] += w.count
            else:
                initial_worker_pool[contractor.id][w.name] = w.count
    cur_worker_pool = deepcopy(initial_worker_pool)

    for sblock in sblocks:
        check_all_allocated_workers_do_not_exceed_capacity_of_contractors(sblock.schedule,
                                                                          initial_worker_pool,
                                                                          cur_worker_pool)



import copy

import numpy as np

from sampo.api.genetic_api import ChromosomeType, ScheduleGenerationScheme
from sampo.scheduler.base import Scheduler
from sampo.scheduler.timeline import JustInTimeTimeline, MomentumTimeline
from sampo.scheduler.timeline.base import Timeline
from sampo.scheduler.timeline.general_timeline import GeneralTimeline
from sampo.scheduler.utils import WorkerContractorPool
from sampo.schemas import ZoneReq
from sampo.schemas.contractor import Contractor
from sampo.schemas.graph import GraphNode
from sampo.schemas.landscape import LandscapeConfiguration
from sampo.schemas.resources import Worker
from sampo.schemas.schedule import ScheduledWork, Schedule
from sampo.schemas.schedule_spec import ScheduleSpec
from sampo.schemas.time import Time
from sampo.schemas.time_estimator import WorkTimeEstimator, DefaultWorkEstimator
from sampo.utilities.linked_list import LinkedList


def convert_schedule_to_chromosome(work_id2index: dict[str, int],
                                   worker_name2index: dict[str, int],
                                   contractor2index: dict[str, int],
                                   contractor_borders: np.ndarray,
                                   schedule: Schedule,
                                   spec: ScheduleSpec,
                                   landscape: LandscapeConfiguration,
                                   order: list[GraphNode] | None = None) -> ChromosomeType:
    """
    Receive a result of scheduling algorithm and transform it to chromosome

    :param work_id2index:
    :param worker_name2index:
    :param contractor2index:
    :param contractor_borders:
    :param schedule:
    :param spec:
    :param landscape:
    :param order: if passed, specify the node order that should appear in the chromosome
    :return:
    """

    order: list[ScheduledWork] = order if order is not None else [work for work in schedule.works
                                                                  if work.id in work_id2index]

    # order works part of chromosome
    order_chromosome: np.ndarray = np.array([work_id2index[work.id] for work in order])

    # convert to convenient form
    schedule = schedule.to_schedule_work_dict

    # resources for works part of chromosome
    # +1 stores contractors line
    resource_chromosome = np.zeros((len(order_chromosome), len(worker_name2index) + 1), dtype=int)

    # zone status changes after node executing
    zone_changes_chromosome = np.zeros((len(order_chromosome), len(landscape.zone_config.start_statuses)), dtype=int)

    for node in order:
        node_id = node.id
        index = work_id2index[node_id]
        for resource in schedule[node_id].workers:
            res_count = resource.count
            res_index = worker_name2index[resource.name]
            res_contractor = resource.contractor_id
            resource_chromosome[index, res_index] = res_count
            resource_chromosome[index, -1] = contractor2index[res_contractor]

    resource_border_chromosome = np.copy(contractor_borders)

    return order_chromosome, resource_chromosome, resource_border_chromosome, spec, zone_changes_chromosome


def convert_chromosome_to_schedule(chromosome: ChromosomeType,
                                   worker_pool: WorkerContractorPool,
                                   index2node: dict[int, GraphNode],
                                   index2contractor: dict[int, Contractor],
                                   index2zone: dict[int, str],
                                   worker_pool_indices: dict[int, dict[int, Worker]],
                                   worker_name2index: dict[str, int],
                                   contractor2index: dict[str, int],
                                   landscape: LandscapeConfiguration,
                                   timeline: Timeline | None = None,
                                   assigned_parent_time: Time = Time(0),
                                   work_estimator: WorkTimeEstimator = DefaultWorkEstimator(),
                                   sgs_type: ScheduleGenerationScheme = ScheduleGenerationScheme.Parallel) \
        -> tuple[dict[GraphNode, ScheduledWork], Time, Timeline, list[GraphNode]]:
    """
    Build schedule from received chromosome
    It can be used in visualization of final solving of genetic algorithm
    """
    match sgs_type:
        case ScheduleGenerationScheme.Parallel:
            converter = parallel_schedule_generation_scheme
        case ScheduleGenerationScheme.Serial:
            converter = serial_schedule_generation_scheme
        case _:
            raise ValueError('Unknown type of schedule generation scheme')
    return converter(chromosome,
                     worker_pool,
                     index2node,
                     index2contractor,
                     index2zone,
                     worker_pool_indices,
                     worker_name2index,
                     contractor2index,
                     landscape,
                     timeline,
                     assigned_parent_time,
                     work_estimator)


def parallel_schedule_generation_scheme(chromosome: ChromosomeType,
                                        worker_pool: WorkerContractorPool,
                                        index2node: dict[int, GraphNode],
                                        index2contractor: dict[int, Contractor],
                                        index2zone: dict[int, str],
                                        worker_pool_indices: dict[int, dict[int, Worker]],
                                        worker_name2index: dict[str, int],
                                        contractor2index: dict[str, int],
                                        landscape: LandscapeConfiguration = LandscapeConfiguration(),
                                        timeline: Timeline | None = None,
                                        assigned_parent_time: Time = Time(0),
                                        work_estimator: WorkTimeEstimator = DefaultWorkEstimator()) \
        -> tuple[dict[GraphNode, ScheduledWork], Time, Timeline, list[GraphNode]]:
    """
    Implementation of Parallel Schedule Generation Scheme
    """
    node2swork: dict[GraphNode, ScheduledWork] = {}

    works_order = chromosome[0]
    works_resources = chromosome[1]
    border = chromosome[2]
    spec = chromosome[3]
    zone_statuses = chromosome[4]
    worker_pool = copy.deepcopy(worker_pool)

    # use 3rd part of chromosome in schedule generator
    for worker_index in worker_pool:
        for contractor_index in worker_pool[worker_index]:
            worker_pool[worker_index][contractor_index].with_count(border[contractor2index[contractor_index],
            worker_name2index[worker_index]])

    if not isinstance(timeline, JustInTimeTimeline):
        timeline = JustInTimeTimeline(worker_pool, landscape)

    order_nodes = []

    # timeline to store starts and ends of all works
    work_timeline = GeneralTimeline()

    def decode(work_index):
        cur_node = index2node[work_index]

        cur_work_spec = spec.get_work_spec(cur_node.id)
        cur_resources = works_resources[work_index, :-1]
        cur_contractor_index = works_resources[work_index, -1]
        cur_contractor = index2contractor[cur_contractor_index]
        cur_worker_team: list[Worker] = [worker_pool_indices[worker_index][cur_contractor_index]
                                         .copy().with_count(worker_count)
                                         for worker_index, worker_count in enumerate(cur_resources)
                                         if worker_count > 0]
        if cur_work_spec.assigned_time is not None:
            cur_exec_time = cur_work_spec.assigned_time
        else:
            cur_exec_time = work_estimator.estimate_time(cur_node.work_unit, cur_worker_team)
        return cur_node, cur_worker_team, cur_contractor, cur_exec_time, cur_work_spec

    # account the remaining works
    enumerated_works_remaining = LinkedList(iterable=enumerate(
        [(work_index, *decode(work_index)) for work_index in works_order]
    ))

    # declare current checkpoint index
    ckpt_idx = 0
    start_time = assigned_parent_time - 1
    prev_start_time = start_time - 1

    def work_scheduled(args) -> bool:
        idx, (work_idx, node, worker_team, contractor, exec_time, work_spec) = args

        if timeline.can_schedule_at_the_moment(node, worker_team, work_spec, node2swork, start_time, exec_time):
            # apply worker spec
            Scheduler.optimize_resources_using_spec(node.work_unit, worker_team, work_spec)

            st = start_time
            if idx == 0:  # we are scheduling the work `start of the project`
                st = assigned_parent_time  # this work should always have st = 0, so we just re-assign it

            if idx == len(works_order) - 1:  # we are scheduling the work `end of the project`
                finish_time, finalizing_zones = timeline.zone_timeline.finish_statuses()
                st = max(start_time, finish_time)

            # finish using time spec
            timeline.schedule(node, node2swork, worker_team, contractor, work_spec,
                              st, exec_time, assigned_parent_time, work_estimator)

            if idx == len(works_order) - 1:  # we are scheduling the work `end of the project`
                node2swork[node].zones_pre = finalizing_zones

            work_timeline.update_timeline(st, exec_time, None)
            return True
        return False

    # while there are unprocessed checkpoints
    while len(enumerated_works_remaining) > 0:
        if ckpt_idx < len(work_timeline):
            start_time = work_timeline[ckpt_idx]
            if prev_start_time == start_time:
                ckpt_idx += 1
                continue
            if start_time.is_inf():
                # break because schedule already contains Time.inf(), that is incorrect schedule
                break
            prev_start_time = start_time
        else:
            start_time += 1

        # find all works that can start at start_time moment and remove it if scheduled
        enumerated_works_remaining.remove_if(work_scheduled)
        ckpt_idx = min(ckpt_idx + 1, len(work_timeline))

    return node2swork, assigned_parent_time, timeline, order_nodes


def serial_schedule_generation_scheme(chromosome: ChromosomeType,
                                      worker_pool: WorkerContractorPool,
                                      index2node: dict[int, GraphNode],
                                      index2contractor: dict[int, Contractor],
                                      index2zone: dict[int, str],
                                      worker_pool_indices: dict[int, dict[int, Worker]],
                                      worker_name2index: dict[str, int],
                                      contractor2index: dict[str, int],
                                      landscape: LandscapeConfiguration = LandscapeConfiguration(),
                                      timeline: Timeline | None = None,
                                      assigned_parent_time: Time = Time(0),
                                      work_estimator: WorkTimeEstimator = DefaultWorkEstimator()) \
        -> tuple[dict[GraphNode, ScheduledWork], Time, Timeline, list[GraphNode]]:
    """
    Implementation of Serial Schedule Generation Scheme
    """
    node2swork: dict[GraphNode, ScheduledWork] = {}

    works_order = chromosome[0]
    works_resources = chromosome[1]
    border = chromosome[2]
    spec = chromosome[3]
    zone_statuses = chromosome[4]
    worker_pool = copy.deepcopy(worker_pool)

    # use 3rd part of chromosome in schedule generator
    for worker_index in worker_pool:
        for contractor_index in worker_pool[worker_index]:
            worker_pool[worker_index][contractor_index].with_count(border[contractor2index[contractor_index],
            worker_name2index[worker_index]])

    if not isinstance(timeline, MomentumTimeline):
        timeline = MomentumTimeline(worker_pool, landscape)

    order_nodes = []

    for order_index, work_index in enumerate(works_order):
        node = index2node[work_index]
        order_nodes.append(node)

        work_spec = spec.get_work_spec(node.id)

        resources = works_resources[work_index, :-1]
        contractor_index = works_resources[work_index, -1]
        contractor = index2contractor[contractor_index]
        worker_team: list[Worker] = [worker_pool[wreq.kind][contractor.id]
                                     .copy().with_count(resources[worker_name2index[wreq.kind]])
                                     for wreq in node.work_unit.worker_reqs]

        # apply worker spec
        Scheduler.optimize_resources_using_spec(node.work_unit, worker_team, work_spec)

        st, ft, exec_times = timeline.find_min_start_time_with_additional(node, worker_team, node2swork, work_spec,
                                                                          assigned_parent_time=assigned_parent_time,
                                                                          work_estimator=work_estimator)

        if order_index == 0:  # we are scheduling the work `start of the project`
            st = assigned_parent_time  # this work should always have st = 0, so we just re-assign it

        # finish using time spec
        ft = timeline.schedule(node, node2swork, worker_team, contractor, work_spec,
                               st, work_spec.assigned_time, assigned_parent_time, work_estimator)
        # process zones
        zone_reqs = [ZoneReq(index2zone[i], zone_status) for i, zone_status in enumerate(zone_statuses[work_index])]
        zone_start_time = timeline.zone_timeline.find_min_start_time(zone_reqs, ft, 0)

        # we should deny scheduling
        # if zone status change can be scheduled only in delayed manner
        if zone_start_time != ft:
            node2swork[node].zones_post = timeline.zone_timeline.update_timeline(order_index,
                                                                                 [z.to_zone() for z in zone_reqs],
                                                                                 zone_start_time, 0)

    schedule_start_time = min((swork.start_time for swork in node2swork.values() if
                               len(swork.workers) != 0), default=assigned_parent_time)

    return node2swork, schedule_start_time, timeline, order_nodes



import random
import numpy as np
from functools import partial
from typing import Type, Callable

from sampo.scheduler.base import SchedulerType
from sampo.scheduler.generic import GenericScheduler, get_finish_time_default
from sampo.scheduler.timeline import Timeline, MomentumTimeline
from sampo.scheduler.utils import WorkerContractorPool, get_worker_contractor_pool
from sampo.schemas.time_estimator import WorkTimeEstimator, DefaultWorkEstimator
from sampo.scheduler.lft.prioritization import lft_prioritization, lft_randomized_prioritization
from sampo.scheduler.lft.time_computaion import work_chain_durations

from sampo.schemas import (Contractor, WorkGraph, GraphNode, LandscapeConfiguration, Worker, Schedule, ScheduledWork,
                           Time, WorkUnit)
from sampo.schemas.schedule_spec import ScheduleSpec, WorkSpec
from sampo.utilities.validation import validate_schedule

from sampo.schemas.exceptions import IncorrectAmountOfWorker, NoSufficientContractorError


def get_contractors_and_workers_amounts_for_work(work_unit: WorkUnit, contractors: list[Contractor],
                                                 spec: ScheduleSpec, worker_pool: WorkerContractorPool) \
        -> tuple[list[Contractor], np.ndarray]:
    """
    This function selects contractors that can perform the work.
    For each selected contractor, the maximum possible amount of workers is assigned,
    if they are not specified in the ScheduleSpec, otherwise the amount from the ScheduleSpec is used.
    """
    work_reqs = work_unit.worker_reqs
    work_spec = spec.get_work_spec(work_unit.id)
    # get assigned amounts of workers in schedule spec
    work_spec_amounts = np.array([work_spec.assigned_workers.get(req.kind, -1) for req in work_reqs])
    # make bool mask of unassigned amounts of workers
    in_spec_mask = work_spec_amounts != -1

    # get min amounts of workers
    min_req_amounts = np.array([req.min_count for req in work_reqs])
    # check validity of assigned in schedule spec amounts of workers
    if (work_spec_amounts[in_spec_mask] < min_req_amounts[in_spec_mask]).any():
        raise IncorrectAmountOfWorker(f"ScheduleSpec assigns not enough workers for work {work_unit.id}")

    # get max amounts of workers
    max_req_amounts = np.array([req.max_count for req in work_reqs])
    # check validity of assigned in schedule spec amounts of workers
    if (work_spec_amounts[in_spec_mask] > max_req_amounts[in_spec_mask]).any():
        raise IncorrectAmountOfWorker(f"ScheduleSpec assigns too many workers for work {work_unit.id}")

    # get contractors borders
    contractors_amounts = np.array([[worker_pool[req.kind][contractor.id].count
                                     if contractor.id in worker_pool[req.kind] else -1
                                     for req in work_reqs]
                                    for contractor in contractors])

    # make bool mask of contractors that satisfy min amounts of workers
    contractors_mask = (contractors_amounts >= min_req_amounts).all(axis=1)
    # update bool mask of contractors to satisfy amounts of workers assigned in schedule spec
    contractors_mask &= (contractors_amounts[:, in_spec_mask] >= work_spec_amounts[in_spec_mask]).all(axis=1)
    # check that there is at least one contractor that satisfies all the constraints
    if not contractors_mask.any():
        raise NoSufficientContractorError(f'There is no contractor that can satisfy given search; contractors: '
                                          f'{contractors}')

    # get contractors that satisfy all the constraints
    accepted_contractors = [contractor for contractor, is_satisfying in zip(contractors, contractors_mask)
                            if is_satisfying]
    if in_spec_mask.all():
        # if all workers are assigned in schedule spec
        # broadcast these amounts on all accepted contractors
        workers_amounts = np.broadcast_to(work_spec_amounts,
                                          (len(accepted_contractors), len(work_spec_amounts)))
    else:
        # if some workers are not assigned in schedule spec
        # then we should assign maximum to them for each contractor
        max_amounts = contractors_amounts[contractors_mask]  # get max amounts of accepted contractors
        # bring max amounts of accepted contractors and max amounts of workers to the same size
        # and take the minimum of them to satisfy all constraints
        max_amounts = np.stack(np.broadcast_arrays(max_amounts, max_req_amounts), axis=0).min(axis=0)
        workers_amounts = max_amounts
        # assign to all accepted contractors assigned in schedule spec amounts of workers
        workers_amounts[:, in_spec_mask] = work_spec_amounts[in_spec_mask]

    return accepted_contractors, workers_amounts


class LFTScheduler(GenericScheduler):
    """
    Scheduler, which assigns contractors evenly, allocates maximum resources
    and schedules works in MIN-LFT priority rule order
    """

    def __init__(self,
                 scheduler_type: SchedulerType = SchedulerType.LFT,
                 timeline_type: Type = MomentumTimeline,
                 work_estimator: WorkTimeEstimator = DefaultWorkEstimator()):
        super().__init__(scheduler_type, None, timeline_type, None, self.get_default_res_opt_function(),
                         work_estimator)
        self.prioritization = lft_prioritization

    def get_default_res_opt_function(self, get_finish_time=get_finish_time_default) \
            -> Callable[[GraphNode, list[Contractor], WorkSpec, WorkerContractorPool,
                         dict[GraphNode, ScheduledWork], Time, Timeline, WorkTimeEstimator],
                        tuple[Time, Time, Contractor, list[Worker]]]:
        def optimize_resources_def(node: GraphNode, contractors: list[Contractor], spec: WorkSpec,
                                   worker_pool: WorkerContractorPool, node2swork: dict[GraphNode, ScheduledWork],
                                   assigned_parent_time: Time, timeline: Timeline, work_estimator: WorkTimeEstimator) \
                -> tuple[Time, Time, Contractor, list[Worker]]:
            # get assigned contractor and workers
            contractor, workers = self._node_id2workers[node.id]
            # find start time
            start_time, finish_time, _ = timeline.find_min_start_time_with_additional(node, workers, node2swork,
                                                                                      spec, None, assigned_parent_time,
                                                                                      work_estimator)
            return start_time, finish_time, contractor, workers

        return optimize_resources_def

    def schedule_with_cache(self,
                            wg: WorkGraph,
                            contractors: list[Contractor],
                            spec: ScheduleSpec = ScheduleSpec(),
                            validate: bool = False,
                            assigned_parent_time: Time = Time(0),
                            timeline: Timeline | None = None,
                            landscape: LandscapeConfiguration() = LandscapeConfiguration()) -> list[tuple[Schedule, Time, Timeline, list[GraphNode]]]:
        # get contractors borders
        worker_pool = get_worker_contractor_pool(contractors)

        # first of all assign workers and contractors to nodes
        # and estimate nodes' durations
        node_id2duration = self._contractor_workers_assignment(wg, contractors, worker_pool, spec)

        # order nodes based on estimated nodes' durations
        ordered_nodes = self.prioritization(wg, node_id2duration)

        if not isinstance(timeline, self._timeline_type):
            timeline = self._timeline_type(worker_pool, landscape)

        # make schedule based on assigned workers, contractors and order
        schedule, schedule_start_time, timeline = self.build_scheduler(ordered_nodes, contractors, landscape, spec,
                                                                       self.work_estimator, assigned_parent_time,
                                                                       timeline)
        del self._node_id2workers
        schedule = Schedule.from_scheduled_works(
            schedule,
            wg
        )

        if validate:
            validate_schedule(schedule, wg, contractors)

        return [(schedule, schedule_start_time, timeline, ordered_nodes)]

    def _contractor_workers_assignment(self, wg: WorkGraph, contractors: list[Contractor],
                                       worker_pool: WorkerContractorPool, spec: ScheduleSpec = ScheduleSpec()
                                       ) -> dict[str, int]:
        # get only heads of chains from work graph nodes
        nodes = [node for node in wg.nodes if not node.is_inseparable_son()]
        # counter for contractors assignments to the works
        contractors_assignments_count = np.ones_like(contractors)
        # mapper of nodes and assigned workers
        self._node_id2workers = {}
        # mapper of nodes and estimated duration
        node_id2duration = {}
        for node in nodes:
            work_unit = node.work_unit
            # get contractors that can perform this work and workers amounts for them
            accepted_contractors, workers_amounts = get_contractors_and_workers_amounts_for_work(work_unit,
                                                                                                 contractors,
                                                                                                 spec,
                                                                                                 worker_pool)

            # estimate chain durations for each accepted contractor
            durations_for_chain = [work_chain_durations(node, amounts, self.work_estimator)
                                   for amounts in workers_amounts]
            # get the sum of the estimated durations for each contractor
            durations = np.array([sum(chain_durations) for chain_durations in durations_for_chain])

            # assign a score for each contractor equal to the sum of the ratios of
            # the duration of this work for this contractor to all durations
            # and the number of assignments of this contractor to the total amount of contractors assignments
            scores = durations / durations.sum() + contractors_assignments_count / contractors_assignments_count.sum()
            # since the maximum possible score value is 2 subtract the resulting scores from 2,
            # so that the higher the score, the more suitable the contractor is for the assignment
            scores = 2 - scores

            # assign contractor based on received scores by implemented strategy
            contractor_index = self._get_contractor_index(scores)
            assigned_contractor = accepted_contractors[contractor_index]

            # get workers amounts of the assigned contractor
            assigned_amount = workers_amounts[contractor_index]

            # increase the counter for the assigned contractor
            contractors_assignments_count[contractor_index] += 1

            # get workers of the assigned contractor and assign them to the node in mapper
            workers = [worker_pool[req.kind][assigned_contractor.id].copy().with_count(amount)
                       for req, amount in zip(work_unit.worker_reqs, assigned_amount)]
            self._node_id2workers[node.id] = (assigned_contractor, workers)

            # assign the received durations to each node in the chain
            for duration, dep_node in zip(durations_for_chain[contractor_index], node.get_inseparable_chain_with_self()):
                node_id2duration[dep_node.id] = duration

        return node_id2duration

    def _get_contractor_index(self, scores: np.ndarray) -> int:
        return np.argmax(scores)


class RandomizedLFTScheduler(LFTScheduler):
    """
    Scheduler, which assigns contractors evenly with stochasticity, allocates maximum resources
    and schedules works in order sampled by MIN-LFT and MIN-LST priority rules
    """

    def __init__(self,
                 scheduler_type: SchedulerType = SchedulerType.LFT,
                 timeline_type: Type = MomentumTimeline,
                 work_estimator: WorkTimeEstimator = DefaultWorkEstimator(),
                 rand: random.Random = random.Random()):
        super().__init__(scheduler_type, timeline_type, work_estimator)
        self._random = rand
        self.prioritization = partial(lft_randomized_prioritization, rand=self._random)

    def _get_contractor_index(self, scores: np.ndarray) -> int:
        return self._random.choices(np.arange(len(scores)), weights=scores)[0] if scores.size > 1 else 0



from collections import defaultdict
from math import ceil

import numpy as np
import torch

from sampo.scheduler.topological.base import TopologicalScheduler
from sampo.schemas.graph import WorkGraph


def one_hot_encode(v, max_v):
    res = [float(0) for _ in range(max_v)]
    res[v] = float(1)
    return res


def one_hot_decode(v: torch.Tensor):
    for i in range(len(v)):
        if v[i] == 1:
            return i


def metric_resource_constrainedness(wg: WorkGraph) -> list[float]:
    """
    The resource constrainedness of a resource type k is defined as  the average number of units requested by all
    activities divided by the capacity of the resource type

    :param wg: Work graph
    :return: List of RC coefficients for each resource type
    """
    resource_dict = defaultdict(lambda: [0, 0])

    for node in wg.nodes:
        for req in node.work_unit.worker_reqs:
            resource_dict[req.kind][0] += 1
            resource_dict[req.kind][1] += req.volume

    return [value[0] / value[1] for name, value in resource_dict.items()]


def metric_graph_parallelism_degree(wg: WorkGraph) -> list[float]:
    batches = 8
    parallelism_degree = []
    current_node = wg.start
    node_count = wg.vertex_count

    stack = [current_node]
    while stack:
        tmp_stack = set()
        parallelism_coef = 0

        for i in range(len(stack)):
            if stack[i] == 0:
                continue
            for j in range(i + 1, len(stack)):
                if stack[j] == 0:
                    continue
                if stack[j] in stack[i].children:
                    stack[j] = 0

        for node in stack:
            if node == 0:
                continue
            parallelism_coef += 1
            for child in node.children:
                tmp_stack.add(child)
        parallelism_degree.append(parallelism_coef / node_count)
        stack = list(tmp_stack)

    step = ceil(len(parallelism_degree) / batches)
    aggregated_degree = [0] * batches
    for i in range(0, len(parallelism_degree), step):
        aggregated_degree[i // step] = np.mean(parallelism_degree[i:(i + step)])

    return aggregated_degree


def metric_longest_path(wg: WorkGraph) -> float:
    scheduler = TopologicalScheduler()
    stack = scheduler.prioritization(wg, None)

    dist = {}
    for node in stack:
        dist[node.id] = 0

    while stack:
        node = stack.pop()
        for child in node.children:
            dist[child.id] = max(dist[child.id], dist[node.id] + 1)

    return max(dist.values())


def metric_vertex_count(wg: WorkGraph) -> float:
    return wg.vertex_count


def metric_average_work_per_activity(wg: WorkGraph) -> float:
    return sum(node.work_unit.volume for node in wg.nodes) / wg.vertex_count


def metric_relative_max_children(wg: WorkGraph) -> float:
    return max((len(node.children) for node in wg.nodes if node.children)) / wg.vertex_count


def metric_average_resource_usage(wg: WorkGraph) -> float:
    return sum(sum((req.min_count + req.max_count) / 2 for req in node.work_unit.worker_reqs)
               for node in wg.nodes) / wg.vertex_count


def metric_relative_max_parents(wg: WorkGraph) -> float:
    return max((len(node.parents) for node in wg.nodes if node.parents)) / wg.vertex_count


def encode_graph(wg: WorkGraph) -> list[float]:
    return [
        metric_vertex_count(wg),
        metric_average_work_per_activity(wg),
        metric_relative_max_children(wg),
        metric_average_resource_usage(wg),
        metric_longest_path(wg),
        *metric_graph_parallelism_degree(wg)
    ]



from collections import deque
from typing import Optional, Union

from sortedcontainers import SortedList

from sampo.scheduler.timeline.base import Timeline
from sampo.scheduler.timeline.to_start_supply_timeline import ToStartSupplyTimeline
from sampo.scheduler.timeline.zone_timeline import ZoneTimeline
from sampo.scheduler.utils import WorkerContractorPool
from sampo.schemas.contractor import Contractor
from sampo.schemas.graph import GraphNode
from sampo.schemas.landscape import LandscapeConfiguration
from sampo.schemas.requirements import WorkerReq
from sampo.schemas.resources import Worker
from sampo.schemas.schedule_spec import WorkSpec
from sampo.schemas.scheduled_work import ScheduledWork
from sampo.schemas.time import Time
from sampo.schemas.time_estimator import WorkTimeEstimator, DefaultWorkEstimator
from sampo.schemas.types import ScheduleEvent, EventType
from sampo.utilities.collections_util import build_index


class MomentumTimeline(Timeline):
    """
    Timeline that stores the intervals in which resources is occupied.
    """

    def __init__(self, worker_pool: WorkerContractorPool, landscape: LandscapeConfiguration, algorithm: str = 'to_start'):
        """
        This should create an empty Timeline from given a list of tasks and contractor list.
        """

        # using  time, seq_id and event_type we can guarantee that
        # there may be only one possible order in cases:
        # (a) when events have the same time
        # (in this cases we need both time and seq_id to properly handle available_workers processing logic)
        # (b) when events have the same time and their start and end matches
        # (service tasks for instance may have zero length)
        def event_cmp(event: Union[ScheduleEvent, Time, tuple[Time, int, int]]) -> tuple[Time, int, int]:
            if isinstance(event, ScheduleEvent):
                if event.event_type is EventType.INITIAL:
                    return Time(-1), -1, event.event_type.priority

                return event.time, event.seq_id, event.event_type.priority

            if isinstance(event, Time):
                # instances of Time must be greater than almost all ScheduleEvents with same time point
                return event, Time.inf().value, 2

            if isinstance(event, tuple):
                return event

            raise ValueError(f'Incorrect type of value: {type(event)}')

        # to efficiently search for time slots for tasks to be scheduled
        # we need to keep track of starts and ends of previously scheduled tasks
        # and remember how many workers of a certain type is available at this particular moment
        self._timeline: dict[str, dict[str, SortedList[ScheduleEvent]]] = {}
        for worker_name, worker_counts in worker_pool.items():
            for contractor, worker in worker_counts.items():
                if contractor not in self._timeline:
                    self._timeline[contractor] = {}
                self._timeline[contractor][worker_name] = SortedList(
                    iterable=(ScheduleEvent(-1, EventType.INITIAL, Time(0), None, worker.count),),
                    key=event_cmp
                )

        # internal index, earlier - task_index parameter for schedule method
        self._task_index = 0
        self._material_timeline = ToStartSupplyTimeline(landscape)
        self.zone_timeline = ZoneTimeline(landscape.zone_config)

    def find_min_start_time_with_additional(self,
                                            node: GraphNode,
                                            worker_team: list[Worker],
                                            node2swork: dict[GraphNode, ScheduledWork],
                                            spec: WorkSpec,
                                            assigned_start_time: Optional[Time] = None,
                                            assigned_parent_time: Time = Time(0),
                                            work_estimator: WorkTimeEstimator = DefaultWorkEstimator()) \
            -> tuple[Time, Time, dict[GraphNode, tuple[Time, Time]]]:
        """
        Looking for an available time slot for given 'GraphNode'

        :param worker_team: list of passed workers. Should be IN THE SAME ORDER AS THE CORRESPONDING WREQS
        :param node: info about given GraphNode
        :param node2swork: dictionary, that match GraphNode to ScheduleWork respectively
        :param spec: schedule specification
        :param assigned_start_time: start time, that can be received from
        another algorithms of calculation the earliest start time
        :param assigned_parent_time: minimum start time
        :param work_estimator: function that calculates execution time of the GraphNode
        :return: start time, end time, time of execution
        """
        inseparable_chain = node.get_inseparable_chain_with_self()

        contractor_id = worker_team[0].contractor_id if worker_team else ''

        # 1. identify earliest possible start time by max parent's end time

        def apply_time_spec(time: Time) -> Time:
            return max(time, assigned_start_time) if assigned_start_time is not None else time

        max_parent_time: Time = max(apply_time_spec(node.min_start_time(node2swork)), assigned_parent_time)

        nodes_max_parent_times = {ins_node: max(apply_time_spec(ins_node.min_start_time(node2swork)),
                                                assigned_parent_time)
                                  for ins_node in inseparable_chain}

        # 2. calculating execution time of the task

        exec_time: Time = Time(0)
        exec_times: dict[GraphNode, tuple[Time, Time]] = {}  # node: (lag, exec_time)
        for chain_node in inseparable_chain:
            node_exec_time: Time = Time(0) if len(chain_node.work_unit.worker_reqs) == 0 else \
                work_estimator.estimate_time(chain_node.work_unit, worker_team)

            lag_req = nodes_max_parent_times[chain_node] - max_parent_time - exec_time
            lag = lag_req if lag_req > 0 else 0

            exec_times[chain_node] = lag, node_exec_time
            exec_time += lag + node_exec_time

        if len(worker_team) == 0:
            max_material_time = self._material_timeline.find_min_material_time(node, max_parent_time,
                                                                               node.work_unit.need_materials())
            max_zone_time = self.zone_timeline.find_min_start_time(node.work_unit.zone_reqs, max_parent_time, exec_time)

            max_parent_time = max(max_parent_time, max_material_time, max_zone_time)
            return max_parent_time, max_parent_time, exec_times

        if assigned_start_time is not None:
            st = assigned_start_time
        else:
            # we can't just use max() of all times we found from different constraints
            # because start time shifting can corrupt time slots we found from every constraint
            # so let's find the time that is agreed with all constraints
            cur_start_time = max_parent_time
            found_earliest_time = False
            while not found_earliest_time:
                cur_start_time = self._find_min_start_time(self._timeline[contractor_id], inseparable_chain, spec,
                                                           cur_start_time, exec_time, worker_team)

                material_time = self._material_timeline.find_min_material_time(node,
                                                                               cur_start_time,
                                                                               node.work_unit.need_materials())
                if material_time > cur_start_time:
                    cur_start_time = material_time
                    continue

                zone_time = self.zone_timeline.find_min_start_time(node.work_unit.zone_reqs, cur_start_time,
                                                                   exec_time)
                if zone_time > cur_start_time:
                    cur_start_time = zone_time
                else:
                    found_earliest_time = True

            st = cur_start_time

        self._validate(st + exec_time, exec_time, worker_team)
        return st, st + exec_time, exec_times

    def _find_min_start_time(self,
                             resource_timeline: dict[str, SortedList[ScheduleEvent]],
                             inseparable_chain: list[GraphNode],
                             spec: WorkSpec,
                             parent_time: Time,
                             exec_time: Time,
                             passed_workers: list[Worker]) -> Time:
        """
        Find start time for the whole 'GraphNode'

        :param resource_timeline: dictionary that stores resource and its Timeline
        :param inseparable_chain: list of GraphNodes that represent one big task,
        that are divided into several dependent tasks
        :param parent_time: the minimum start time
        :param exec_time: the time of execution
        :param passed_workers: list of passed workers. Should be IN THE SAME ORDER AS THE CORRESPONDING WREQS
        :return:
        """
        # if it is a service unit, then it can be satisfied by any contractor at any moment
        # because no real workers are going to be used to execute the task,
        # however, we still should respect dependencies of the service task
        # and should start it only after all the dependency tasks are done
        if all((node.work_unit.is_service_unit for node in inseparable_chain)):
            return parent_time

        # checking if the contractor can satisfy requirements for the task at all,
        # we return None in cases when the task cannot be executed
        # even if it is scheduled to the very end, e.g., after the end of all other tasks
        # already scheduled to this contractor

        for node in inseparable_chain:
            for i, wreq in enumerate(node.work_unit.worker_reqs):
                initial_event: ScheduleEvent = resource_timeline[wreq.kind][0]
                assert initial_event.event_type is EventType.INITIAL
                # if this contractor initially has fewer workers of this type, then needed...
                if initial_event.available_workers_count < passed_workers[i].count:
                    return Time.inf()

        # here we look for the earliest time slot that can satisfy all the worker's specializations
        # we do it in that manner because each worker specialization can be treated separately
        # e.g. requested for different tasks
        # We check only the first node since all inseparable nodes have
        # same worker_reqs despite the difference in exec time
        queue = deque(inseparable_chain[0].work_unit.worker_reqs)

        start = parent_time
        scheduled_wreqs: list[WorkerReq] = []

        type2count: dict[str, int] = build_index(passed_workers, lambda w: w.name, lambda w: w.count)

        i = 0
        while len(queue) > 0:

            i += 1

            wreq = queue.popleft()
            state = resource_timeline[wreq.kind]
            # we look for the earliest time slot starting from 'start' time moment
            # if we have found a time slot for the previous task,
            # we should start to find for the earliest time slot of other task since this new time
            found_start = self._find_earliest_time_slot(state, start, exec_time, type2count[wreq.kind], spec)

            assert found_start >= start

            if scheduled_wreqs and start != found_start:
                # The current worker specialization can be started only later than
                # the previously found start time.
                # In this case we need to add back all previously scheduled wreq-s into the queue
                # to be scheduled again with the new start time (e.g. found start).
                # This process should reach its termination at least at the very end of this contractor's schedule.
                queue.extend(scheduled_wreqs)
                scheduled_wreqs.clear()

            scheduled_wreqs.append(wreq)
            start = found_start

        return start

    @staticmethod
    def _find_earliest_time_slot(state: SortedList[ScheduleEvent],
                                 parent_time: Time,
                                 exec_time: Time,
                                 required_worker_count: int,
                                 spec: WorkSpec) -> Time:
        """
        Searches for the earliest time starting from start_time, when a time slot
        of exec_time is available, when required_worker_count of resources is available

        :param state: stores Timeline for the certain resource
        :param parent_time: the minimum start time starting from the end of the parent task
        :param exec_time: execution time of work
        :param required_worker_count: requirements amount of Worker
        :return: the earliest start time
        """
        # if the work has zero execution time, then there is no need to take resources
        if exec_time == 0:
            return parent_time

        current_start_time = parent_time
        current_start_idx = state.bisect_right(current_start_time) - 1

        last_time = state[-1].time

        if spec.is_independent:
            return max(parent_time, last_time)

        # the condition means we have reached the end of schedule for this contractor subject to specialization (wreq)
        # as long as we assured that this contractor has enough capacity at all to handle the task
        # we can stop and put the task at the very end
        while current_start_time < last_time:

            end_idx = state.bisect_left((current_start_time + exec_time, -1, EventType.INITIAL))

            # checking from the end of execution interval, i.e., end_idx - 1
            # up to (including) the event right prepending or equal the start
            # of the execution interval, i.e., current_start_idx
            not_enough_workers_found = False
            for idx in range(end_idx - 1, current_start_idx - 1, -1):
                if state[idx].available_workers_count < required_worker_count:
                    # we're trying to find a new slot that would start with
                    # the last index passing the quantity check
                    current_start_idx = idx + 1
                    not_enough_workers_found = True
                    break

            if not not_enough_workers_found:
                break

            current_start_time = state[current_start_idx].time

        return current_start_time

    def can_schedule_at_the_moment(self,
                                   node: GraphNode,
                                   worker_team: list[Worker],
                                   spec: WorkSpec,
                                   node2swork: dict[GraphNode, ScheduledWork],
                                   start_time: Time,
                                   exec_time: Time) -> bool:
        if spec.is_independent:
            # squash all the timeline to the last point
            for worker in worker_team:
                worker_timeline = self._timeline[(worker.contractor_id, worker.name)]
                last_cpkt_time, _ = worker_timeline[0]
                if last_cpkt_time >= start_time:
                    return False
            return True
        else:
            start = start_time
            end = start_time + exec_time

            # checking availability of renewable resources
            for w in worker_team:
                state = self._timeline[w.contractor_id][w.name]
                start_idx = state.bisect_right(start)
                end_idx = state.bisect_left((end, -1, EventType.INITIAL))
                available_workers_count = state[start_idx - 1].available_workers_count
                # updating all events in between the start and the end of our current task
                for event in state[start_idx: end_idx]:
                    if not event.available_workers_count >= w.count:
                        return False

                if not available_workers_count >= w.count:
                    return False

                if not state[0].available_workers_count >= available_workers_count:
                    return False

            if not self._material_timeline.can_schedule_at_the_moment(node, start_time,
                                                                      node.work_unit.need_materials()):
                return False
            if not self.zone_timeline.can_schedule_at_the_moment(node.work_unit.zone_reqs, start_time, exec_time):
                return False

            return True

    def update_timeline(self,
                        finish_time: Time,
                        exec_time: Time,
                        node: GraphNode,
                        worker_team: list[Worker],
                        spec: WorkSpec):
        """
        Inserts `chosen_workers` into the timeline with it's `inseparable_chain`
        """
        # if the work has zero execution time, then there is no need to take resources
        if exec_time == 0:
            return
        # 7. for each worker's specialization of the chosen contractor being used by the task
        # we update counts of available workers on previously scheduled events
        # that lie between start and end of the task
        # Also, add events of the start and the end to worker's specializations
        # of the chosen contractor.

        task_index = self._task_index
        self._task_index += 1

        # experimental logics lightening. debugging showed its efficiency.

        start = finish_time - exec_time
        end = finish_time
        for w in worker_team:
            state = self._timeline[w.contractor_id][w.name]
            start_idx = state.bisect_right(start)
            end_idx = state.bisect_left((end, -1, EventType.INITIAL))
            available_workers_count = state[start_idx - 1].available_workers_count
            # updating all events in between the start and the end of our current task
            for event in state[start_idx: end_idx]:
                assert event.available_workers_count >= w.count
                event.available_workers_count -= w.count

            assert available_workers_count >= w.count

            state.add(ScheduleEvent(task_index, EventType.START, start, None, available_workers_count - w.count))

            # move the index on time when resources of the work will be freed
            end_idx = state.bisect_right(end) - 1
            if state[end_idx].time == end:
                # time when resources will be freed is already in timeline
                # so resources amount will be the same at the moment
                end_count = state[end_idx].available_workers_count
            else:
                # time when resources will be freed is not in timeline
                end_count = state[end_idx].available_workers_count + w.count

            state.add(ScheduleEvent(task_index, EventType.END, end, None, end_count))

    def schedule(self,
                 node: GraphNode,
                 node2swork: dict[GraphNode, ScheduledWork],
                 workers: list[Worker],
                 contractor: Contractor,
                 spec: WorkSpec,
                 assigned_start_time: Optional[Time] = None,
                 assigned_time: Optional[Time] = None,
                 assigned_parent_time: Time = Time(0),
                 work_estimator: WorkTimeEstimator = DefaultWorkEstimator()):
        inseparable_chain = node.get_inseparable_chain_with_self()
        start_time, _, exec_times = \
            self.find_min_start_time_with_additional(node, workers, node2swork, spec, assigned_start_time,
                                                     assigned_parent_time, work_estimator)
        if assigned_time is not None:
            exec_times = {n: (Time(0), assigned_time // len(inseparable_chain))
                          for n in inseparable_chain}

        # TODO Decide how to deal with exec_times(maybe we should remove using pre-computed exec_times)
        self._schedule_with_inseparables(node, node2swork, inseparable_chain, spec,
                                         workers, contractor, start_time, exec_times)

    def _schedule_with_inseparables(self,
                                    node: GraphNode,
                                    node2swork: dict[GraphNode, ScheduledWork],
                                    inseparable_chain: list[GraphNode],
                                    spec: WorkSpec,
                                    worker_team: list[Worker],
                                    contractor: Contractor,
                                    start_time: Time,
                                    exec_times: dict[GraphNode, tuple[Time, Time]]):
        # 6. create a schedule entry for the task
        # nodes_start_times = {ins_node: ins_node.min_start_time(node2swork) for ins_node in inseparable_chain}

        curr_time = start_time
        for i, chain_node in enumerate(inseparable_chain):
            node_lag, node_time = exec_times[chain_node]

            # lag_req = nodes_start_times[chain_node] - curr_time
            # node_lag = lag_req if lag_req > 0 else 0

            start_work = curr_time + node_lag
            deliveries, mat_del_time = self._material_timeline.deliver_resources(chain_node,
                                                                                 start_work,
                                                                                 chain_node.work_unit.need_materials(),
                                                                                 True)
            start_work = max(start_work, mat_del_time)
            self._validate(start_work + node_time, node_time, worker_team)
            swork = ScheduledWork(
                work_unit=chain_node.work_unit,
                start_end_time=(start_work, start_work + node_time),
                workers=worker_team,
                contractor=contractor,
                materials=deliveries
            )
            curr_time = start_work + node_time
            node2swork[chain_node] = swork

        self.update_timeline(curr_time, curr_time - start_time, node, worker_team, spec)
        zones = [zone_req.to_zone() for zone_req in node.work_unit.zone_reqs]
        node2swork[node].zones_pre = self.zone_timeline.update_timeline(len(node2swork), zones, start_time,
                                                                        curr_time - start_time)

    def _validate(self,
                  finish_time: Time,
                  exec_time: Time,
                  worker_team: list[Worker]):
        if exec_time == 0:
            return

        start = finish_time - exec_time
        end = finish_time
        for w in worker_team:
            state = self._timeline[w.contractor_id][w.name]
            start_idx = state.bisect_right(start)
            end_idx = state.bisect_left((end, -1, EventType.INITIAL))
            available_workers_count = state[start_idx - 1].available_workers_count
            # updating all events in between the start and the end of our current task
            for event in state[start_idx: end_idx]:
                assert event.available_workers_count >= w.count

            assert available_workers_count >= w.count



import random

import numpy as np
from deap.base import Toolbox

from sampo.api.genetic_api import ChromosomeType
from sampo.scheduler.genetic.converter import convert_schedule_to_chromosome, ScheduleGenerationScheme
from sampo.scheduler.genetic.operators import init_toolbox
from sampo.scheduler.utils import get_worker_contractor_pool
from sampo.schemas import WorkGraph, Contractor, Schedule, GraphNode, LandscapeConfiguration, WorkTimeEstimator, Time
from sampo.schemas.schedule_spec import ScheduleSpec


def init_chromosomes_f(wg: WorkGraph,
                       contractors: list[Contractor],
                       init_schedules: dict[str, tuple[Schedule, list[GraphNode] | None, ScheduleSpec, float]],
                       landscape: LandscapeConfiguration = LandscapeConfiguration()):
    worker_pool, index2node, index2zone, work_id2index, worker_name2index, index2contractor_obj, \
        worker_pool_indices, contractor2index, contractor_borders, node_indices, parents, children, \
        resources_border = prepare_optimized_data_structures(wg, contractors, landscape)

    init_chromosomes: dict[str, tuple[ChromosomeType, float, ScheduleSpec]] = \
        {name: (convert_schedule_to_chromosome(work_id2index, worker_name2index,
                                               contractor2index, contractor_borders, schedule, chromosome_spec,
                                               landscape, order),
                importance, chromosome_spec)
         if schedule is not None else None
         for name, (schedule, order, chromosome_spec, importance) in init_schedules.items()}

    return init_chromosomes


def prepare_optimized_data_structures(wg: WorkGraph,
                                      contractors: list[Contractor],
                                      landscape: LandscapeConfiguration):
    # preparing access-optimized data structures
    nodes = [node for node in wg.nodes if not node.is_inseparable_son()]

    worker_pool = get_worker_contractor_pool(contractors)

    index2node: dict[int, GraphNode] = {index: node for index, node in enumerate(nodes)}
    work_id2index: dict[str, int] = {node.id: index for index, node in index2node.items()}
    worker_name2index = {worker_name: index for index, worker_name in enumerate(worker_pool)}
    index2contractor_obj = {ind: contractor for ind, contractor in enumerate(contractors)}
    index2zone = {ind: zone for ind, zone in enumerate(landscape.zone_config.start_statuses)}
    contractor2index = {contractor.id: ind for ind, contractor in enumerate(contractors)}
    worker_pool_indices = {worker_name2index[worker_name]: {
        contractor2index[contractor_id]: worker for contractor_id, worker in workers_of_type.items()
    } for worker_name, workers_of_type in worker_pool.items()}
    node_indices = list(range(len(nodes)))

    # construct inseparable_child -> inseparable_parent mapping
    inseparable_parents = {}
    for node in nodes:
        for child in node.get_inseparable_chain_with_self():
            inseparable_parents[child] = node

    # here we aggregate information about relationships from the whole inseparable chain
    children = {work_id2index[node.id]: set([work_id2index[inseparable_parents[child].id]
                                             for inseparable in node.get_inseparable_chain_with_self()
                                             for child in inseparable.children])
                for node in nodes}

    parents = {work_id2index[node.id]: set() for node in nodes}
    for node, node_children in children.items():
        for child in node_children:
            parents[child].add(node)

    resources_border = np.zeros((2, len(worker_pool), len(index2node)))
    for work_index, node in index2node.items():
        for req in node.work_unit.worker_reqs:
            worker_index = worker_name2index[req.kind]
            resources_border[0, worker_index, work_index] = req.min_count
            resources_border[1, worker_index, work_index] = req.max_count

    contractor_borders = np.zeros((len(contractor2index), len(worker_name2index)), dtype=int)
    for ind, contractor in enumerate(contractors):
        for ind_worker, worker in enumerate(contractor.workers.values()):
            contractor_borders[ind, ind_worker] = worker.count

    return (worker_pool, index2node, index2zone, work_id2index, worker_name2index, index2contractor_obj,
            worker_pool_indices, contractor2index, contractor_borders, node_indices, parents, children,
            resources_border)


def create_toolbox_using_cached_chromosomes(wg: WorkGraph,
                                            contractors: list[Contractor],
                                            population_size: int,
                                            mutate_order: float,
                                            mutate_resources: float,
                                            mutate_zones: float,
                                            init_chromosomes: dict[str, tuple[ChromosomeType, float, ScheduleSpec]],
                                            rand: random.Random,
                                            spec: ScheduleSpec = ScheduleSpec(),
                                            work_estimator: WorkTimeEstimator = None,
                                            assigned_parent_time: Time = Time(0),
                                            fitness_weights: tuple[int | float, ...] = (-1,),
                                            landscape: LandscapeConfiguration = LandscapeConfiguration(),
                                            sgs_type: ScheduleGenerationScheme = ScheduleGenerationScheme.Parallel,
                                            only_lft_initialization: bool = False,
                                            is_multiobjective: bool = False) -> Toolbox:
    worker_pool, index2node, index2zone, work_id2index, worker_name2index, index2contractor_obj, \
        worker_pool_indices, contractor2index, contractor_borders, node_indices, parents, children, \
        resources_border = prepare_optimized_data_structures(wg, contractors, landscape)

    return init_toolbox(wg,
                        contractors,
                        worker_pool,
                        landscape,
                        index2node,
                        work_id2index,
                        worker_name2index,
                        index2contractor_obj,
                        index2zone,
                        init_chromosomes,
                        mutate_order,
                        mutate_resources,
                        mutate_zones,
                        landscape.zone_config.statuses.statuses_available(),
                        population_size,
                        rand,
                        spec,
                        worker_pool_indices,
                        contractor2index,
                        contractor_borders,
                        node_indices,
                        parents,
                        children,
                        resources_border,
                        assigned_parent_time,
                        fitness_weights,
                        work_estimator,
                        sgs_type,
                        only_lft_initialization,
                        is_multiobjective)



from typing import Callable, Optional

import numpy as np

from sampo.schemas.time import Time

"""
Contains base optimization methods
"""


# float version
def dichotomy_float(down: float, up: float, func, eps: float = 0.000001) -> float:
    delta = eps / 5
    while up - down > eps:
        x1 = (down + up - delta) / 2
        x2 = (down + up + delta) / 2
        if func(x1) < func(x2):
            up = x2
        else:
            down = x1

    return down


# int version
def dichotomy_int(down: int, up: int, func: Callable[[int], Time]) -> int:
    while up - down > 2:
        x1 = (down + up - 1) >> 1
        x2 = (down + up + 1) >> 1
        if x1 == x2:
            return x1
        if func(x1) < func(x2):
            up = x2
        else:
            down = x1

        # print(str(x1) + ' ' + str(x2) + ' ' + str(down) + ' ' + str(up))

    return (up + down) >> 1


def coordinate_descent(down: np.ndarray, up: np.ndarray,
                       method: Callable[[int, int, Callable[[int], Time]], Time],
                       fitness: Callable[[np.ndarray], Time],
                       optimize_array: Optional[np.ndarray]) -> np.ndarray:
    cur = down.copy()
    for i in range(down.size):
        if optimize_array and not optimize_array[i]:
            continue

        def part(x) -> Time:
            cur[i] = x
            return fitness(cur)

        cur[i] = method(down[i], up[i], part)
    return cur



from sampo.schemas.requirements import WorkerReq
from sampo.schemas.time import Time


def scale_reqs(req_list: list[WorkerReq], scalar: float, new_name: str | None = None) -> list[WorkerReq]:
    """
    Multiplies each of the requirements of their list, scaling the maximum number of resources and volume

    :param req_list: list of resource requirements
    :param scalar: scalar to which the requirements are applied
    :param new_name: A new name for the requirements
    :return: scaled requirements
    """
    return [work_req.scale_all(scalar, new_name) for work_req in req_list]


def mul_volume_reqs(req_list: list[WorkerReq], scalar: float, new_name: str | None = None) -> list[WorkerReq]:
    """
    Multiplies each of the requirements of their list, scaling only the volume

    :param req_list: list of resource requirements
    :param scalar: scalar to which the requirements are applied
    :param new_name: A new name for the requirements
    :return: scaled requirements
        """
    return [work_req.scale_volume(scalar, new_name) for work_req in req_list]


def get_borehole_volume(borehole_count: int, base: (float, float)) -> float:
    """
    Function to calculate the scalar for objects depending on the number of boreholes

    :param borehole_count: number of boreholes on the site
    :param base:
        base[0] part of the volume independent of the number of boreholes,
        base[1] part of the volume dependent on the number of boreholes
    :return: returns a scalar to calculate the requirements
    """
    return base[0] + base[1] * borehole_count


def mul_borehole_volume(req_list: list[WorkerReq], borehole_count: int, base: (float, float)) -> list[WorkerReq]:
    """
    Function for scaling resource requirements for works dependent on the number of boreholes

    :param req_list: list of resource requirements
    :param borehole_count: number of boreholes on the site
    :param base:
        base[0] part of the volume independent of the number of boreholes,
        base[1] part of the volume dependent on the number of boreholes
    :return: scaled requirements
    """
    return mul_volume_reqs(req_list, get_borehole_volume(borehole_count, base))


START_PROJECT = []
END_PROJECT = []


# <--road_block-->
# scaled by: scale_reqs
# measurement base: 1 km
MIN_ROAD = [
    WorkerReq(kind='driver', volume=Time(15), min_count=3,
              max_count=9, name='min_road'),
    WorkerReq(kind='manager', volume=Time(5), min_count=1,
              max_count=3, name='min_road'),
    WorkerReq(kind='handyman', volume=Time(30), min_count=6,
              max_count=18, name='min_road'),
]
ATOMIC_ROAD_LEN = 0.05
TEMP_ROAD = mul_volume_reqs(MIN_ROAD, scalar=4, new_name='temp_road')
MIN_TEMP_ROAD = mul_volume_reqs(MIN_ROAD, scalar=5, new_name='temp_road')
FINAL_ROAD = mul_volume_reqs(MIN_ROAD, scalar=10, new_name='final_road')
# /<--road_block-->


# <--engineering_preparation-->
# scaled by: mul_borehole_volume
# measurement base: tuple(volume_without_boreholes, volume_for_one_borehole)
ENGINEERING_PREPARATION_BASE = (0.44, 0.03)
ENGINEERING_PREPARATION = [
    WorkerReq(kind='driver', volume=Time(900), min_count=6,
              max_count=30, name='engineering_preparation'),
    WorkerReq(kind='manager', volume=Time(150), min_count=2,
              max_count=6, name='engineering_preparation'),
    WorkerReq(kind='handyman', volume=Time(900), min_count=6,
              max_count=30, name='engineering_preparation'),
    WorkerReq(kind='engineer', volume=Time(150), min_count=2,
              max_count=6, name='engineering_preparation'),
]
# /<--engineering_preparation-->


# <--power_line-->
# scaled by: scale_reqs
# measurement base: 1 km
POWER_LINE = [
    WorkerReq(kind='driver', volume=Time(120), min_count=6,
              max_count=9, name='power_line'),
    WorkerReq(kind='fitter', volume=Time(120), min_count=6,
              max_count=9, name='power_line'),
    WorkerReq(kind='manager', volume=Time(40), min_count=2,
              max_count=6, name='power_line'),
    WorkerReq(kind='handyman', volume=Time(120), min_count=6,
              max_count=9, name='power_line'),
    WorkerReq(kind='electrician', volume=Time(40), min_count=2,
              max_count=6, name='power_line'),
]
HIGH_POWER_LINE = mul_volume_reqs(POWER_LINE, scalar=1.5, new_name='high_power_line')
# /<--power_line-->

# <--pipe_line-->
# scaled by: scale_reqs
# measurement base: 1 km
PIPE_LINE = scale_reqs(POWER_LINE, scalar=1, new_name='pipe_line')
LOOPING = mul_volume_reqs(PIPE_LINE, scalar=0.8, new_name='looping')
# /<--pipe_line-->
# <--pipe_node-->
# scaled by: -
# measurement base: -
PIPE_NODE = [
    WorkerReq(kind='driver', volume=Time(400), min_count=2,
              max_count=6, name='pipe_node'),
    WorkerReq(kind='fitter', volume=Time(800), min_count=4,
              max_count=8, name='pipe_node'),
    WorkerReq(kind='manager', volume=Time(400), min_count=2,
              max_count=8, name='pipe_node'),
    WorkerReq(kind='handyman', volume=Time(1200), min_count=6,
              max_count=12, name='pipe_node'),
    WorkerReq(kind='electrician', volume=Time(300), min_count=2,
              max_count=6, name='pipe_node'),
]
# /<--pipe_node-->


# <--metering_install-->
# scaled by: -
# measurement base: -
METERING_INSTALL = [
    WorkerReq(kind='driver', volume=Time(200), min_count=2,
              max_count=6, name='metering_install'),
    WorkerReq(kind='fitter', volume=Time(300), min_count=3,
              max_count=9, name='metering_install'),
    WorkerReq(kind='manager', volume=Time(50), min_count=1,
              max_count=2, name='metering_install'),
    WorkerReq(kind='handyman', volume=Time(300), min_count=4,
              max_count=12, name='metering_install'),
    WorkerReq(kind='engineer', volume=Time(100), min_count=2,
              max_count=4, name='metering_install'),
]
# /<--metering_install-->
# <--ktp_nep-->
# scaled by: mul_borehole_volume
# measurement base: tuple(volume_without_boreholes, volume_for_one_borehole)
KTP_NEP_BASE = (0.58, 0.03)
KTP_NEP = [
    WorkerReq(kind='driver', volume=Time(200), min_count=2,
              max_count=6, name='ktp_nep'),
    WorkerReq(kind='fitter', volume=Time(300), min_count=3,
              max_count=9, name='ktp_nep'),
    WorkerReq(kind='manager', volume=Time(50), min_count=1,
              max_count=2, name='ktp_nep'),
    WorkerReq(kind='handyman', volume=Time(300), min_count=4,
              max_count=12, name='ktp_nep'),
    WorkerReq(kind='electrician', volume=Time(100), min_count=2,
              max_count=4, name='ktp_nep'),
]
# <--ktp_nep-->
# scaled by: mul_volume_reqs
# measurement base: DRAINAGE_TANK_BASE_VOLUME + DRAINAGE_TANK_INPUT_VOLUME * borehole_count
# scaled by: mul_borehole_volume
# measurement base: tuple(volume_without_boreholes, volume_for_one_borehole)
DRAINAGE_TANK_BASE = (0.3, 0.05)
DRAINAGE_TANK = [
    WorkerReq(kind='driver', volume=Time(400), min_count=4,
              max_count=8, name='drainage_tank'),
    WorkerReq(kind='manager', volume=Time(50), min_count=1,
              max_count=2, name='drainage_tank'),
    WorkerReq(kind='handyman', volume=Time(300), min_count=3,
              max_count=9, name='drainage_tank'),
]

# <--boreholes_equipment_shared-->
# scaled by: -
# measurement base: -
WATER_BLOCK = scale_reqs(METERING_INSTALL, scalar=1.1, new_name='water_block')
AUTOMATION_BLOCK = scale_reqs(METERING_INSTALL, scalar=0.9, new_name='automation_block')
BLOCK_DOSAGE = scale_reqs(METERING_INSTALL, scalar=1, new_name='block_dosage')
START_FILTER = scale_reqs(METERING_INSTALL, scalar=1.05, new_name='start_filter_system')
FIREWALL = scale_reqs(DRAINAGE_TANK, scalar=1.05, new_name='firewall_tank')
# /<--boreholes_equipment_shared-->


# <--borehole-->
# scaled by: -
# measurement base: -
BOREHOLE = [
    WorkerReq(kind='driver', volume=Time(200), min_count=4,
              max_count=12, name='borehole'),
    WorkerReq(kind='fitter', volume=Time(150), min_count=3,
              max_count=9, name='borehole'),
    WorkerReq(kind='manager', volume=Time(25), min_count=1,
              max_count=3, name='borehole'),
    WorkerReq(kind='handyman', volume=Time(250), min_count=6,
              max_count=9, name='borehole')
]
# /<--borehole-->


# <--pipe_and_power_network-->
# scaled by: scale_reqs
# measurement base: 1 km
PIPE_NETWORK = [
    WorkerReq(kind='driver', volume=Time(60), min_count=3,
              max_count=6, name='pipe_network'),
    WorkerReq(kind='fitter', volume=Time(60), min_count=3,
              max_count=6, name='pipe_network'),
    WorkerReq(kind='manager', volume=Time(20), min_count=2,
              max_count=4, name='pipe_network'),
    WorkerReq(kind='handyman', volume=Time(50), min_count=3,
              max_count=6, name='pipe_network'),
    WorkerReq(kind='electrician', volume=Time(20), min_count=2,
              max_count=4, name='pipe_network'),
]
POWER_NETWORK = scale_reqs(PIPE_NETWORK, scalar=1, new_name='power_network')
# /<--pipe_and_power_network-->
# <--flooding_light_mast-->
# scaled by: -
# measurement base: -
LIGHT_MAST = [
    WorkerReq(kind='driver', volume=Time(20), min_count=2,
              max_count=4, name='flooding_light_mast'),
    WorkerReq(kind='fitter', volume=Time(40), min_count=4,
              max_count=8, name='flooding_light_mast'),
    WorkerReq(kind='manager', volume=Time(10), min_count=2,
              max_count=4, name='flooding_light_mast'),
    WorkerReq(kind='handyman', volume=Time(60), min_count=3,
              max_count=12, name='flooding_light_mast'),
    WorkerReq(kind='electrician', volume=Time(5), min_count=1,
              max_count=1, name='flooding_light_mast'),
]
# /<--flooding_light_mast-->

# scaled by: mul_borehole_volume
# measurement base: tuple(volume_without_boreholes, volume_for_one_borehole)
HANDING_STAGE_BASE = (0.66, 0.02)
HANDING_STAGE = [
    WorkerReq(kind='driver', volume=Time(5), min_count=1,
              max_count=3, name='handing_stage'),
    WorkerReq(kind='manager', volume=Time(5), min_count=1,
              max_count=3, name='handing_stage'),
    WorkerReq(kind='engineer', volume=Time(10), min_count=2,
              max_count=8, name='handing_stage'),
]



import pandas as pd

from sampo.generator.environment import ContractorGenerationMethod
from sampo.pipeline.base import InputPipeline, SchedulePipeline
from sampo.pipeline.delegating import DelegatingScheduler
from sampo.pipeline.lag_optimization import LagOptimizationStrategy
from sampo.pipeline.preparation import PreparationPipeline
from sampo.scheduler.base import Scheduler
from sampo.scheduler.generic import GenericScheduler
from sampo.scheduler.utils import get_worker_contractor_pool
from sampo.scheduler.utils.local_optimization import OrderLocalOptimizer, ScheduleLocalOptimizer
from sampo.schemas.apply_queue import ApplyQueue
from sampo.schemas.contractor import Contractor
from sampo.schemas.exceptions import NoSufficientContractorError
from sampo.schemas.graph import WorkGraph, GraphNode
from sampo.schemas.landscape import LandscapeConfiguration
from sampo.schemas.project import ScheduledProject
from sampo.schemas.schedule import Schedule
from sampo.schemas.schedule_spec import ScheduleSpec
from sampo.schemas.time import Time
from sampo.schemas.time_estimator import WorkTimeEstimator, DefaultWorkEstimator
from sampo.structurator import graph_restructuring
from sampo.userinput.parser.csv_parser import CSVParser
from sampo.utilities.name_mapper import NameMapper, read_json


def contractors_can_perform_work_graph(contractors: list[Contractor], wg: WorkGraph) -> bool:
    is_at_least_one_contractor_can_perform = True
    num_contractors_can_perform_node = 0

    for node in wg.nodes:
        reqs = node.work_unit.worker_reqs
        for contractor in contractors:
            offers = contractor.workers
            for req in reqs:
                if req.min_count > offers[req.kind].count:
                    is_at_least_one_contractor_can_perform = False
                    break
            if is_at_least_one_contractor_can_perform:
                num_contractors_can_perform_node += 1
                break
            is_at_least_one_contractor_can_perform = True
        if num_contractors_can_perform_node == 0:
            return False
        num_contractors_can_perform_node = 0

    return True


class DefaultInputPipeline(InputPipeline):
    """
    Default pipeline, that help to use the framework
    """

    def __init__(self):
        self._wg: WorkGraph | pd.DataFrame | str | None = None
        self._contractors: list[Contractor] | pd.DataFrame | str | tuple[ContractorGenerationMethod, int] | None \
            = ContractorGenerationMethod.AVG, 1
        self._work_estimator: WorkTimeEstimator = DefaultWorkEstimator()
        self._node_orders: list[list[GraphNode]] | None = None
        self._lag_optimize: LagOptimizationStrategy = LagOptimizationStrategy.NONE
        self._spec: ScheduleSpec | None = ScheduleSpec()
        self._assigned_parent_time: Time | None = Time(0)
        self._local_optimize_stack: ApplyQueue = ApplyQueue()
        self._landscape_config = LandscapeConfiguration()
        self._preparation = PreparationPipeline()
        self._history: pd.DataFrame = pd.DataFrame(columns=['marker_for_glue', 'work_name', 'first_day', 'last_day',
                                                            'upper_works', 'work_name_clear_old', 'smr_name',
                                                            'work_name_clear', 'granular_smr_name'])
        self._all_connections: bool = False
        self._change_connections_info: bool = False
        self._name_mapper: NameMapper | None = None
        self.sep_wg = ';'
        self.sep_history = ';'

    def wg(self,
           wg: WorkGraph | pd.DataFrame | str,
           change_base_on_history: bool = False,
           sep: str = ';',
           all_connections: bool = False,
           change_connections_info: bool = False) -> 'InputPipeline':
        """
        Mandatory argument.

        :param change_base_on_history: whether it is necessary to change project information based on connection history data
        :param is_wg_has_full_info_about_connections: does the project information contain full details of the works
        :param wg: the WorkGraph object for scheduling task
        :param sep: separating character. It's mandatory, if you send the file path with work_info

        ATTENTION!
            If you send WorkGraph .csv or HistoryData file path, use the same separating character in
            work_info.csv as in history_data.csv and vice versa.
        """
        self._wg = wg
        self._all_connections = all_connections
        self._change_connections_info = change_connections_info
        self.sep_wg = sep
        return self

    def contractors(self, contractors: list[Contractor] | pd.DataFrame | str | tuple[ContractorGenerationMethod, int]) \
            -> 'InputPipeline':
        """
        Mandatory argument.

        :param contractors: the contractors list for scheduling task, or DataFrame with contractor info,
                            or file with contractor info, of method for contractors generation with
                            number of contractors to be generated
        :return: the pipeline object
        """
        self._contractors = contractors
        return self

    def landscape(self, landscape_config: LandscapeConfiguration) -> 'InputPipeline':
        """
        Set landscape configuration

        :param landscape_config:
        :return:
        """
        self._landscape_config = landscape_config
        return self

    def name_mapper(self, name_mapper: NameMapper | str) -> 'InputPipeline':
        """
        Set works' name mapper
        :param name_mapper:
        :return:
        """
        if isinstance(name_mapper, str):
            name_mapper = read_json(name_mapper)
        self._name_mapper = name_mapper
        return self

    def history(self, history: pd.DataFrame | str, sep: str = ';') -> 'InputPipeline':
        """
        Set historical data. Mandatory method, if work graph hasn't info about links
        :param history:
        :param sep: separating character. It's mandatory, if you send the file path with work_info

        ATTENTION!
            If you send WorkGraph .csv or HistoryData file path, use the same separating character in
            work_info.csv as in history_data.csv and vice versa.
        """
        self._history = history
        self.sep_history = sep
        return self

    def spec(self, spec: ScheduleSpec) -> 'InputPipeline':
        """
        Set specification of schedule

        :param spec:
        :return:
        """
        self._spec = spec
        return self

    def time_shift(self, time: Time) -> 'InputPipeline':
        """
        If the schedule should start at a certain time

        :param time:
        :return:
        """
        self._assigned_parent_time = time
        return self

    def lag_optimize(self, lag_optimize: LagOptimizationStrategy) -> 'InputPipeline':
        """
        Mandatory argument. Shows should graph be lag-optimized or not.
        If not defined, pipeline should search the best variant of this argument in result.

        :param lag_optimize:
        :return: the pipeline object
        """
        self._lag_optimize = lag_optimize
        return self

    def work_estimator(self, work_estimator: WorkTimeEstimator) -> 'InputPipeline':
        self._work_estimator = work_estimator
        return self

    def node_order(self, node_orders: list[list[GraphNode]]) -> 'InputPipeline':
        self._node_orders = node_orders
        return self

    def optimize_local(self, optimizer: OrderLocalOptimizer, area: range) -> 'InputPipeline':
        self._local_optimize_stack.add(optimizer.optimize, area)
        return self

    def schedule(self, scheduler: Scheduler) -> 'SchedulePipeline':
        if isinstance(self._wg, pd.DataFrame) or isinstance(self._wg, str):
            self._wg, self._contractors = \
                CSVParser.work_graph_and_contractors(
                    works_info=CSVParser.read_graph_info(project_info=self._wg,
                                                         history_data=self._history,
                                                         sep_wg=self.sep_wg,
                                                         sep_history=self.sep_history,
                                                         name_mapper=self._name_mapper,
                                                         all_connections=self._all_connections,
                                                         change_connections_info=self._change_connections_info),
                    contractor_info=self._contractors,
                    work_resource_estimator=self._work_estimator
                )

        if not contractors_can_perform_work_graph(self._contractors, self._wg):
            raise NoSufficientContractorError('Contractors are not able to perform the graph of works')

        if isinstance(scheduler, GenericScheduler):
            # if scheduler is generic, it supports injecting local optimizations
            # cache upper-layer self to another variable to get it from inner class
            s_self = self

            class LocalOptimizedScheduler(DelegatingScheduler):

                def __init__(self, delegate: GenericScheduler):
                    super().__init__(delegate)

                def delegate_prioritization(self, orig_prioritization):
                    def prioritization(wg: WorkGraph, work_estimator: WorkTimeEstimator):
                        # call delegate's prioritization and apply local optimizations
                        return s_self._local_optimize_stack.apply(orig_prioritization(wg, work_estimator))

                    return prioritization

            scheduler = LocalOptimizedScheduler(scheduler)
        elif not self._local_optimize_stack.empty():
            print('Trying to apply local optimizations to non-generic scheduler, ignoring it')

        match self._lag_optimize:
            case LagOptimizationStrategy.NONE:
                wg = self._wg
                schedules = scheduler.schedule_with_cache(wg, self._contractors,
                                                          self._spec,
                                                          landscape=self._landscape_config,
                                                          assigned_parent_time=self._assigned_parent_time)
                node_orders = [node_order for _, _, _, node_order in schedules]
                schedules = [schedule for schedule, _, _, _ in schedules]
                self._node_orders = node_orders

            case LagOptimizationStrategy.AUTO:
                # Searching the best
                wg1 = graph_restructuring(self._wg, False)
                schedules = scheduler.schedule_with_cache(wg1, self._contractors,
                                                          self._spec,
                                                          landscape=self._landscape_config,
                                                          assigned_parent_time=self._assigned_parent_time)
                node_orders1 = [node_order for _, _, _, node_order in schedules]
                schedules1 = [schedule for schedule, _, _, _ in schedules]
                min_time1 = min([schedule.execution_time for schedule in schedules1])

                wg2 = graph_restructuring(self._wg, True)
                schedules = scheduler.schedule_with_cache(wg2, self._contractors,
                                                          self._spec,
                                                          landscape=self._landscape_config,
                                                          assigned_parent_time=self._assigned_parent_time)
                node_orders2 = [node_order for _, _, _, node_order in schedules]
                schedules2 = [schedule for schedule, _, _, _ in schedules]
                min_time2 = min([schedule.execution_time for schedule in schedules2])

                if min_time1 < min_time2:
                    self._node_orders = node_orders1
                    wg = wg1
                    schedules = schedules1
                else:
                    self._node_orders = node_orders2
                    wg = wg2
                    schedules = schedules2

            case _:
                wg = graph_restructuring(self._wg, self._lag_optimize.value)
                schedules = scheduler.schedule_with_cache(wg, self._contractors,
                                                          self._spec,
                                                          landscape=self._landscape_config,
                                                          assigned_parent_time=self._assigned_parent_time)
                node_orders = [node_order for _, _, _, node_order in schedules]
                schedules = [schedule for schedule, _, _, _ in schedules]
                self._node_orders = node_orders

        return DefaultSchedulePipeline(self, wg, schedules)


# noinspection PyProtectedMember
class DefaultSchedulePipeline(SchedulePipeline):

    def __init__(self, s_input: DefaultInputPipeline, wg: WorkGraph, schedules: list[Schedule]):
        self._input = s_input
        self._wg = wg
        self._worker_pool = get_worker_contractor_pool(s_input._contractors)
        self._schedules = schedules
        self._scheduled_works = [{wg[swork.id]: swork for swork in schedule.to_schedule_work_dict.values()}
                                 for schedule in schedules]
        self._local_optimize_stack = ApplyQueue()
        self._start_date = None

    def optimize_local(self, optimizer: ScheduleLocalOptimizer, area: range) -> 'SchedulePipeline':
        self._local_optimize_stack.add(optimizer.optimize,
                                       self._input._contractors, self._input._landscape_config,
                                       self._input._spec, self._worker_pool, self._input._work_estimator,
                                       self._input._assigned_parent_time, area)
        return self

    def finish(self) -> list[ScheduledProject]:
        scheduled_projects = []
        for scheduled_works, node_order in zip(self._scheduled_works, self._input._node_orders):
            processed_sworks = self._local_optimize_stack.apply(scheduled_works, node_order)
            schedule = Schedule.from_scheduled_works(processed_sworks.values(), self._wg)
            scheduled_projects.append(ScheduledProject(self._input._wg, self._wg, self._input._contractors, schedule))
        return scheduled_projects

    def visualization(self, start_date: str) -> list['Visualization']:
        from sampo.utilities.visualization import Visualization
        return [Visualization.from_project(project, start_date) for project in self.finish()]



from typing import Callable

import numpy as np

from sampo.scheduler.utils import WorkerContractorPool
from sampo.schemas.contractor import Contractor
from sampo.schemas.exceptions import NoSufficientContractorError
from sampo.schemas.requirements import WorkerReq
from sampo.schemas.resources import Worker
from sampo.schemas.time import Time


def get_worker_borders(agents: WorkerContractorPool, contractor: Contractor, work_reqs: list[WorkerReq]) \
        -> tuple[np.ndarray, np.ndarray, list[Worker]]:
    """
    Define for each job each type of workers the min and max possible number of workers.
    For max number of workers, max is defined as a minimum from max possible numbers
    at all and max possible for a current job.
    
    :param agents: from all projects
    :param contractor:
    :param work_reqs:
    :return:
    """
    n = len(work_reqs)
    min_worker_team = np.zeros(n, dtype=int)
    max_worker_team = np.zeros(n, dtype=int)
    workers: list[Worker] = []

    for i, req in enumerate(work_reqs):
        w = agents[req.kind].get(contractor.id, None)

        if w is None or w.count < req.min_count:
            # can't satisfy requirements, return empty to say it
            return [], [], []

        min_worker_team[i] = req.min_count
        max_worker_team[i] = min(req.max_count, w.count)
        workers.append(w)

    return min_worker_team, max_worker_team, workers


def run_contractor_search(contractors: list[Contractor],
                          runner: Callable[[Contractor], tuple[Time, Time, list[Worker]]]) \
        -> tuple[Time, Time, Contractor, list[Worker]]:
    """
    Performs the best contractor search.
    
    :param contractors: contractors' list
    :param runner: a runner function, should be inner of the calling code.
        Calculates Tuple[start time, finish time, worker team] from given contractor object.
    :return: start time, finish time, the best contractor, worker team with the best contractor
    """
    # TODO Parallelize

    # optimization metric
    best_finish_time = Time.inf()
    best_contractor = None
    # heuristic: if contractors' finish times are equal, we prefer smaller one
    best_contractor_size = float('inf')

    for contractor in contractors:
        start_time, finish_time, worker_team = runner(contractor)
        contractor_size = sum(w.count for w in contractor.workers.values())

        if not finish_time.is_inf() and (finish_time < best_finish_time or
                                         (finish_time == best_finish_time and contractor_size < best_contractor_size)):
            best_finish_time = finish_time
            best_contractor = contractor
            best_contractor_size = contractor_size

    if best_contractor is None:
        raise NoSufficientContractorError(f'There is no contractor that can satisfy given search; contractors: '
                                          f'{contractors}')

    best_start_time, best_finish_time, best_worker_team = runner(best_contractor)

    return best_start_time, best_finish_time, best_contractor, best_worker_team



import random
from typing import Optional

from sampo.api.genetic_api import ChromosomeType
from sampo.scheduler.base import Scheduler, SchedulerType
from sampo.scheduler.genetic.operators import FitnessFunction, TimeFitness
from sampo.scheduler.genetic.schedule_builder import build_schedules
from sampo.scheduler.genetic.converter import ScheduleGenerationScheme
from sampo.scheduler.heft.base import HEFTScheduler, HEFTBetweenScheduler
from sampo.scheduler.lft.base import LFTScheduler
from sampo.scheduler.resource.average_req import AverageReqResourceOptimizer
from sampo.scheduler.resource.base import ResourceOptimizer
from sampo.scheduler.resource.identity import IdentityResourceOptimizer
from sampo.scheduler.resources_in_time.average_binary_search import AverageBinarySearchResourceOptimizingScheduler
from sampo.scheduler.timeline.base import Timeline
from sampo.schemas.contractor import Contractor
from sampo.schemas.exceptions import NoSufficientContractorError
from sampo.schemas.graph import WorkGraph, GraphNode
from sampo.schemas.landscape import LandscapeConfiguration
from sampo.schemas.schedule import Schedule
from sampo.schemas.schedule_spec import ScheduleSpec
from sampo.schemas.time import Time
from sampo.schemas.time_estimator import WorkTimeEstimator, DefaultWorkEstimator
from sampo.utilities.validation import validate_schedule


class GeneticScheduler(Scheduler):
    """
    Class for hybrid scheduling algorithm, that uses heuristic algorithm to generate
    first population and genetic algorithm to search the best solving
    """

    def __init__(self,
                 number_of_generation: Optional[int] = 50,
                 mutate_order: Optional[float or None] = None,
                 mutate_resources: Optional[float or None] = None,
                 mutate_zones: Optional[float or None] = None,
                 size_of_population: Optional[float or None] = None,
                 rand: Optional[random.Random] = None,
                 seed: Optional[float or None] = None,
                 weights: Optional[list[int] or None] = None,
                 fitness_constructor: FitnessFunction = TimeFitness(),
                 fitness_weights: tuple[int | float, ...] = (-1,),
                 scheduler_type: SchedulerType = SchedulerType.Genetic,
                 resource_optimizer: ResourceOptimizer = IdentityResourceOptimizer(),
                 work_estimator: WorkTimeEstimator = DefaultWorkEstimator(),
                 sgs_type: ScheduleGenerationScheme = ScheduleGenerationScheme.Parallel,
                 optimize_resources: bool = False,
                 is_multiobjective: bool = False,
                 only_lft_initialization: bool = False):
        super().__init__(scheduler_type=scheduler_type,
                         resource_optimizer=resource_optimizer,
                         work_estimator=work_estimator)
        self.number_of_generation = number_of_generation
        self.mutate_order = mutate_order
        self.mutate_resources = mutate_resources
        self.mutate_zones = mutate_zones
        self.size_of_population = size_of_population
        self.rand = rand or random.Random(seed)
        self.fitness_constructor = fitness_constructor
        self.fitness_weights = fitness_weights
        self.work_estimator = work_estimator
        self.sgs_type = sgs_type

        self._optimize_resources = optimize_resources
        self._is_multiobjective = is_multiobjective
        self._weights = weights
        self._only_lft_initialization = only_lft_initialization

        self._time_border = None
        self._max_plateau_steps = None
        self._deadline = None

    def __str__(self) -> str:
        return f'GeneticScheduler[' \
               f'generations={self.number_of_generation},' \
               f'population_size={self.size_of_population},' \
               f'mutate_order={self.mutate_order},' \
               f'mutate_resources={self.mutate_resources}' \
               f']'

    def get_params(self, works_count: int) -> tuple[float, float, float, int]:
        """
        Return base parameters for model to make new population

        :param works_count:
        :return:
        """
        mutate_order = self.mutate_order
        if mutate_order is None:
            mutate_order = 0.05

        mutate_resources = self.mutate_resources
        if mutate_resources is None:
            mutate_resources = 0.05

        mutate_zones = self.mutate_zones
        if mutate_zones is None:
            mutate_zones = 0.05

        size_of_population = self.size_of_population
        if size_of_population is None:
            if works_count < 300:
                size_of_population = 50
            elif 1500 > works_count >= 300:
                size_of_population = 100
            else:
                size_of_population = works_count // 25
        return mutate_order, mutate_resources, mutate_zones, size_of_population

    def set_time_border(self, time_border: int):
        self._time_border = time_border

    def set_max_plateau_steps(self, max_plateau_steps: int):
        self._max_plateau_steps = max_plateau_steps

    def set_deadline(self, deadline: Time):
        """
        Set the deadline of tasks

        :param deadline:
        """
        self._deadline = deadline

    def set_weights(self, weights: list[int]):
        self._weights = weights

    def set_optimize_resources(self, optimize_resources: bool):
        self._optimize_resources = optimize_resources

    def set_is_multiobjective(self, is_multiobjective: bool):
        self._is_multiobjective = is_multiobjective

    def set_only_lft_initialization(self, only_lft_initialization: bool):
        self._only_lft_initialization = only_lft_initialization

    @staticmethod
    def generate_first_population(wg: WorkGraph,
                                  contractors: list[Contractor],
                                  landscape: LandscapeConfiguration = LandscapeConfiguration(),
                                  spec: ScheduleSpec = ScheduleSpec(),
                                  work_estimator: WorkTimeEstimator = None,
                                  deadline: Time = None,
                                  weights=None):
        """
        Algorithm, that generate first population

        :param landscape:
        :param wg: graph of works
        :param contractors:
        :param spec:
        :param work_estimator:
        :param deadline:
        :param weights:
        :return:
        """

        if weights is None:
            weights = [2, 2, 2, 1, 1, 1, 1]

        schedule, _, _, node_order = LFTScheduler(work_estimator=work_estimator).schedule_with_cache(wg, contractors,
                                                                                                     spec,
                                                                                                     landscape=landscape)[0]
        init_lft_schedule = (schedule, node_order[::-1], spec)

        def init_k_schedule(scheduler_class, k) -> tuple[Schedule | None, list[GraphNode] | None, ScheduleSpec | None]:
            try:
                schedule, _, _, node_order = (scheduler_class(work_estimator=work_estimator,
                                                              resource_optimizer=AverageReqResourceOptimizer(k))
                                              .schedule_with_cache(wg, contractors, spec, landscape=landscape))[0]
                return schedule, node_order[::-1], spec
            except NoSufficientContractorError:
                return None, None, None

        if deadline is None:
            def init_schedule(scheduler_class) -> tuple[Schedule | None, list[GraphNode] | None, ScheduleSpec | None]:
                try:
                    schedule, _, _, node_order = (scheduler_class(work_estimator=work_estimator)
                                                  .schedule_with_cache(wg, contractors, spec, landscape=landscape))[0]
                    return schedule, node_order[::-1], spec
                except NoSufficientContractorError:
                    return None, None, None

        else:
            def init_schedule(scheduler_class) -> tuple[Schedule | None, list[GraphNode] | None, ScheduleSpec | None]:
                try:
                    (schedule, _, _, node_order), modified_spec = AverageBinarySearchResourceOptimizingScheduler(
                        scheduler_class(work_estimator=work_estimator)
                    ).schedule_with_cache(wg, contractors, deadline, spec, landscape=landscape)
                    return schedule, node_order[::-1], modified_spec
                except NoSufficientContractorError:
                    return None, None, None

        return {
            "lft": (*init_lft_schedule, weights[0]),
            "heft_end": (*init_schedule(HEFTScheduler), weights[1]),
            "heft_between": (*init_schedule(HEFTBetweenScheduler), weights[2]),
            "12.5%": (*init_k_schedule(HEFTScheduler, 8), weights[3]),
            "25%": (*init_k_schedule(HEFTScheduler, 4), weights[4]),
            "75%": (*init_k_schedule(HEFTScheduler, 4 / 3), weights[5]),
            "87.5%": (*init_k_schedule(HEFTScheduler, 8 / 7), weights[6])
        }

    def schedule_with_cache(self,
                            wg: WorkGraph,
                            contractors: list[Contractor],
                            spec: ScheduleSpec = ScheduleSpec(),
                            validate: bool = False,
                            assigned_parent_time: Time = Time(0),
                            timeline: Timeline | None = None,
                            landscape: LandscapeConfiguration = LandscapeConfiguration()) \
            -> list[tuple[Schedule, Time, Timeline, list[GraphNode]]]:
        """
        Build schedules for received graph of workers and return the current state of schedules
        It's needed to use this method in multy agents model

        :param landscape:
        :param wg:
        :param contractors:
        :param spec:
        :param validate:
        :param assigned_parent_time:
        :param timeline:
        :return:
        """
        init_schedules = GeneticScheduler.generate_first_population(wg, contractors, landscape, spec,
                                                                    self.work_estimator, self._deadline, self._weights)

        mutate_order, mutate_resources, mutate_zones, size_of_population = self.get_params(wg.vertex_count)
        deadline = None if self._optimize_resources else self._deadline

        schedules = build_schedules(wg,
                                    contractors,
                                    size_of_population,
                                    self.number_of_generation,
                                    mutate_order,
                                    mutate_resources,
                                    mutate_zones,
                                    init_schedules,
                                    self.rand,
                                    spec,
                                    self._weights,
                                    landscape,
                                    self.fitness_constructor,
                                    self.fitness_weights,
                                    self.work_estimator,
                                    self.sgs_type,
                                    assigned_parent_time,
                                    timeline,
                                    self._time_border,
                                    self._max_plateau_steps,
                                    self._optimize_resources,
                                    deadline,
                                    self._only_lft_initialization,
                                    self._is_multiobjective)
        schedules = [
            (Schedule.from_scheduled_works(scheduled_works.values(), wg), schedule_start_time, timeline, order_nodes)
            for scheduled_works, schedule_start_time, timeline, order_nodes in schedules]

        if validate:
            for schedule, *_ in schedules:
                validate_schedule(schedule, wg, contractors)

        return schedules



from uuid import uuid4

import numpy as np

from sampo.schemas import GraphNode, Worker, WorkTimeEstimator


def work_chain_durations(node: GraphNode, assigned_workers_amounts: np.ndarray, work_estimator: WorkTimeEstimator) \
        -> list[int]:
    work_unit = node.work_unit

    passed_workers = [Worker(str(uuid4()), req.kind, assigned_amount)
                      for req, assigned_amount in zip(work_unit.worker_reqs, assigned_workers_amounts)]

    chain_durations = [work_estimator.estimate_time(dep_node.work_unit, passed_workers).value + 1
                       for dep_node in node.get_inseparable_chain_with_self()]

    return chain_durations



from sampo.scheduler.multi_agency.block_generator import SyntheticBlockGraphType
from sampo.scheduler.multi_agency.block_graph import BlockGraph, BlockNode
from sampo.scheduler.multi_agency.block_validation import validate_block_schedule
from sampo.scheduler.multi_agency.multi_agency import Agent, Manager, NeuralManager, NoSufficientAgents



import math
from collections import defaultdict

from sortedcontainers import SortedList

from sampo.scheduler.timeline.base import BaseSupplyTimeline
from sampo.scheduler.timeline.platform_timeline import PlatformTimeline
from sampo.schemas.exceptions import NotEnoughMaterialsInDepots, NoDepots, NoAvailableResources
from sampo.schemas.graph import GraphNode
from sampo.schemas.landscape import LandscapeConfiguration, ResourceHolder, Vehicle, Road, MaterialDelivery
from sampo.schemas.landscape_graph import LandGraphNode
from sampo.schemas.resources import Material
from sampo.schemas.time import Time
from sampo.schemas.types import ScheduleEvent, EventType


class ToStartSupplyTimeline(BaseSupplyTimeline):
    def __init__(self, landscape_config: LandscapeConfiguration):

        def event_cmp(event: ScheduleEvent | Time | tuple[Time, int, int]) -> tuple[Time, int, int]:
            if isinstance(event, ScheduleEvent):
                if event.event_type is EventType.INITIAL:
                    return Time(-1), -1, event.event_type.priority

                return event.time, event.seq_id, event.event_type.priority

            if isinstance(event, Time):
                # instances of Time must be greater than almost all ScheduleEvents with same time point
                return event, Time.inf().value, 2

            if isinstance(event, tuple):
                return event

            raise ValueError(f'Incorrect type of value: {type(event)}')

        self._platform_timeline = PlatformTimeline(landscape_config)
        self._timeline: dict[str, dict[str, SortedList[ScheduleEvent]]] = {}
        self._task_index = 0
        self._node_id2holder: dict[str, ResourceHolder] = landscape_config.holder_node_id2resource_holder
        self._holder_id2holder: dict[str, ResourceHolder] = {holder.id: holder for holder in landscape_config.holders}
        self._landscape = landscape_config
        for resource in landscape_config.get_all_resources():
            for mat_id, mat_dict in resource.items():
                self._timeline[mat_id] = {
                    mat[0]: SortedList(iterable=(ScheduleEvent(-1, EventType.INITIAL, Time(0), None, mat[1]),),
                                       key=event_cmp)
                    for mat in mat_dict.items()
                }

    @staticmethod
    def _get_vehicles_need(depot: ResourceHolder, materials: list[Material]) -> int:
        vehicle_capacity = depot.vehicles[0].capacity
        need_mat = {mat.name: mat.count for mat in materials}
        return max(math.ceil(need_mat[material_carry_one_vehicle.name] / material_carry_one_vehicle.count)
                   for material_carry_one_vehicle in vehicle_capacity
                   if material_carry_one_vehicle.name in need_mat)

    @staticmethod
    def _check_resource_availability(state: SortedList[ScheduleEvent],
                                     required_resources: int,
                                     start_ind: int,
                                     finish_ind: int) -> bool:
        for idx in range(finish_ind - 1, start_ind - 1, -1):
            if state[idx].available_workers_count < required_resources:
                return False
        return True

    def _can_deliver_to_time(self, node: GraphNode, finish_delivery_time: Time, materials: list[Material]) -> bool:
        _, time = self._supply_resources(node, finish_delivery_time, materials)
        assert time >= finish_delivery_time
        return time == finish_delivery_time

    def can_schedule_at_the_moment(self, node: GraphNode, start_time: Time,
                                   materials: list[Material]) -> bool:
        if not node.work_unit.need_materials():
            return True

        if not self._platform_timeline.can_schedule_at_the_moment(node, start_time, materials):
            return False

        materials_for_delivery = self._platform_timeline.get_material_for_delivery(node, materials, start_time)
        if sum((mat.count for mat in materials_for_delivery), 0) == 0:
            # there are no materials to be delivered
            return True

        can_delivery = self._can_deliver_to_time(node, start_time, materials_for_delivery)
        return can_delivery

    def _check_material_availability_on_platform(self, platform: LandGraphNode, materials: list[Material],
                                                 start_time: Time) -> bool:
        """
        Check the materials' availability on the `platform` at the `start_time`
        """
        # TODO Add delivery opportunity checking
        platform_state = self._timeline[platform.id]

        for mat in materials:
            start = platform_state[mat.name].bisect_right(start_time) - 1
            finish = platform_state[mat.name].bisect_left((Time.inf(), -1, EventType.INITIAL)) - 1

            if finish - start > 1:
                return False

        return True

    def find_min_material_time(self, node: GraphNode, start_time: Time,
                               materials: list[Material]) -> Time:
        if node.work_unit.is_service_unit or not materials:
            return start_time

        start_time, mat_request = self._platform_timeline.find_min_material_time_with_additional(node, start_time, materials)
        if sum(mat.count for mat in mat_request) == 0:
            return start_time

        _, time = self._supply_resources(node, start_time, materials)
        return time

    def _find_best_holders_by_dist(self, node: LandGraphNode,
                                   materials: list[Material]) -> list[str]:
        """
        Get the depot, that is the closest and the earliest resource-supply available.
        :param node: GraphNode, that initializes resource delivery
        :param materials: required materials
        :return: best depot, time
        """
        sorted_holder_ids = [self._node_id2holder[holder_id].id
                             for dist, holder_id in self._landscape.get_sorted_holders(node)]

        if not sorted_holder_ids:
            raise NoDepots(f'Schedule can not be built. There is no any resource holder')

        holders_can_supply_materials = []
        for holder_id in sorted_holder_ids:
            holder_state = self._timeline[holder_id]
            materials_available = 0

            for material in materials:
                if holder_state.get(material.name, None) is not None:
                    # TODO Check bisect_left
                    ind = holder_state[material.name].bisect_left((Time.inf(), -1, EventType.INITIAL)) - 1
                    if holder_state[material.name][ind].available_workers_count >= material.count:
                        materials_available += 1
            if materials_available == len(materials):
                holders_can_supply_materials.append(holder_id)

        if not holders_can_supply_materials:
            raise NotEnoughMaterialsInDepots(
                f'Schedule can not be built. There is no resource holder that has required materials')

        return holders_can_supply_materials

    def deliver_resources(self,
                          node: GraphNode,
                          deadline: Time,
                          materials: list[Material],
                          update: bool = False) -> tuple[MaterialDelivery, Time]:
        if not node.work_unit.need_materials():
            return MaterialDelivery(node.id), deadline

        if self._platform_timeline.can_provide_resources(node, deadline, materials, update):
            return MaterialDelivery(node.id), deadline

        materials_for_delivery = self._platform_timeline.get_material_for_delivery(node, materials, deadline)
        delivery, time = self._supply_resources(node, deadline, materials_for_delivery, update)

        # print(node.id)
        return delivery, time

    def _supply_resources(self, node: GraphNode,
                          deadline: Time,
                          materials: list[Material],
                          update: bool = False) -> tuple[MaterialDelivery, Time]:
        """
        Finds minimal time that given materials can be supplied, greater than given start time

        :param update: should timeline be updated
        It is necessary when materials are supplied, otherwise, timeline should not be updated
        :param node: GraphNode that initializes the resource-delivery
        :param deadline: the time work starts
        :param materials: material resources that are required to start
        :return: material deliveries, the time when resources are ready
        """
        delivery = MaterialDelivery(node.id)

        if not materials:
            return delivery, deadline

        platform = self._landscape.works2platform[node]

        # get the best depot that has enough materials and get its access start time (absolute value)
        depots = [self._holder_id2holder[depot] for depot in self._find_best_holders_by_dist(platform, materials)]
        selected_depot = None
        # TODO Rename
        min_depot_time = Time.inf()

        selected_vehicles = []
        depot_vehicle_finish_time = Time(0)

        start_delivery_time = Time(-1)
        finish_delivery_time = Time(-1)

        road_deliveries = []

        # find time, that depot could provide resources
        while finish_delivery_time < deadline:
            start_delivery_time += 1
            # min start time found on this iteration
            local_min_start_time = Time.inf()

            for depot in depots:
                depot_mat_start_time = start_delivery_time
                # TODO Check that is necessary
                # for mat in materials:
                #     depot_mat_start_time = max(self._find_earliest_start_time(self._timeline[depot][mat.name],
                #                                                               mat.count,
                #                                                               depot_mat_start_time,
                #                                                               Time.inf()), depot_mat_start_time)

                # get vehicles from the depot
                vehicle_count_need = self._get_vehicles_need(depot, materials)
                vehicle_state = self._timeline[depot.id]['vehicles']
                selected_vehicles = depot.vehicles[:vehicle_count_need]

                route_start_time, deliveries, exec_ahead_time, exec_return_time = \
                    self._get_route_info(depot.node, platform, selected_vehicles, depot_mat_start_time)

                depot_vehicle_start_time = self._find_earliest_start_time(vehicle_state,
                                                                          vehicle_count_need,
                                                                          route_start_time,
                                                                          exec_return_time + exec_ahead_time)


                if local_min_start_time > depot_vehicle_start_time:
                    depot_vehicle_finish_time = depot_vehicle_start_time + exec_return_time + exec_ahead_time
                    # FIXME min_depot_time and local_min_start_time have the same value when while ends
                    min_depot_time = depot_vehicle_start_time
                    local_min_start_time = depot_vehicle_start_time
                    finish_delivery_time = depot_vehicle_start_time + exec_ahead_time
                    selected_depot = depot

                    road_deliveries = deliveries

        for mat in materials:
            delivery.add_delivery(mat.name, mat.count, min_depot_time, finish_delivery_time, selected_depot.name)

        if not update:
            return delivery, finish_delivery_time

        update_timeline_info: dict[str, list[tuple[str, int, Time, Time]]] = defaultdict(list)

        # add update info about holder
        update_timeline_info[selected_depot.id] = [(mat.name, mat.count, min_depot_time, min_depot_time + Time.inf())
                                                   for mat in materials]
        update_timeline_info[selected_depot.id].append(('vehicles', len(selected_vehicles),
                                                        min_depot_time, depot_vehicle_finish_time))

        # add update info about roads
        for delivery_dict in road_deliveries:
            for road_id, res_info in delivery_dict.items():
                update_timeline_info[road_id].append(res_info)

        node_mat_req = {mat.name: mat.count for mat in node.work_unit.need_materials()}
        update_mat_req_info = [(mat.name, node_mat_req[mat.name] - mat.count, finish_delivery_time) for mat in materials]
        self._platform_timeline.update_timeline(platform.id, update_mat_req_info)

        self._update_timeline(update_timeline_info)

        return delivery, finish_delivery_time

    @staticmethod
    def _find_earliest_start_time(state: SortedList[ScheduleEvent],
                                  required_resources: int,
                                  parent_time: Time,
                                  exec_time: Time) -> Time:
        current_start_time = parent_time
        base_ind = state.bisect_right(parent_time) - 1

        ind = state.bisect_left((Time.inf(), -1, EventType)) - 1
        last_time = state[ind].time

        while current_start_time < last_time:
            end_ind = state.bisect_right(current_start_time + exec_time)

            not_enough_resources = False
            for idx in range(end_ind - 1, base_ind - 1, -1):
                if state[idx].available_workers_count < required_resources:
                    base_ind = min(len(state) - 1, idx + 1)
                    not_enough_resources = True
                    break

            if not not_enough_resources:
                break

            current_start_time = state[base_ind].time

        assert current_start_time >= parent_time

        return current_start_time

    def _get_route_info(self, holder_node: LandGraphNode, node: LandGraphNode, vehicles: list[Vehicle],
                        start_holder_time: Time) -> tuple[Time, list[dict[str, tuple[str, int, Time, Time]]], Time, Time]:
        def move_vehicles(from_node: LandGraphNode,
                          to_node: LandGraphNode,
                          parent_time: Time,
                          batch_size: int):
            road_delivery: dict[str, tuple[str, int, Time, Time]] = {}
            finish_time = parent_time

            available_roads = [road for road in self._landscape.roads if road.vehicles >= batch_size]
            route = self._landscape.construct_route(from_node, to_node, available_roads)

            if not route:
                raise NoAvailableResources(f'there is no chance to construct route with roads '
                                           f'{[road.name for road in available_roads]}')

            # check time availability of each part of 'route'
            for road_id in route:
                road_overcome_time = _id2road[road_id].overcome_time
                for i in range(len(vehicles) // max(batch_size, 1)):
                    start_time = self._find_earliest_start_time(state=self._timeline[road_id]['vehicles'],
                                                                required_resources=batch_size,
                                                                parent_time=finish_time,
                                                                exec_time=road_overcome_time)
                    road_delivery[road_id] = ('vehicles', batch_size,
                                              start_time, start_time + Time(road_overcome_time))
                    finish_time = start_time + road_overcome_time

            if not road_delivery:
                raise NoAvailableResources(f'there is no resources of available roads '
                                           f'{[road.name for road in available_roads]} '
                                           f'(probably roads have less bandwidth than is required)')

            return finish_time, road_delivery, finish_time - parent_time

        _id2road: dict[str, Road] = {road.id: road for road in self._landscape.roads}

        # | ------------ from holder to platform ------------ |
        finish_delivery_time, delivery, exec_time_ahead = move_vehicles(holder_node, node, start_holder_time,
                                                                        len(vehicles))

        # | ------------ from platform to holder ------------ |
        # compute the return time for vehicles
        # TODO: adapt move_vehicles() for batch_size = 1. Now the method don't allow to save deliveries of each (batch_size = 1) vehicle (similar with roads' deliveries)
        return_time, return_delivery, exec_time_return = move_vehicles(node, holder_node, finish_delivery_time,
                                                                       len(vehicles))

        return finish_delivery_time - exec_time_ahead, [delivery, return_delivery], exec_time_ahead, exec_time_return

    def _update_timeline(self, update_timeline_info: dict[str, list[tuple[str, int, Time, Time]]]) -> None:
        for res_holder_id, res_holder_info in update_timeline_info.items():
            res_holder_state = self._timeline[res_holder_id]

            for res_info in res_holder_info:
                task_index = self._task_index
                self._task_index += 1

                res_name, res_count, start_time, end_time = res_info
                res_state = res_holder_state[res_name]
                start_idx = res_state.bisect_right(start_time)

                end_idx = res_state.bisect_left((end_time, -1, EventType.INITIAL))
                available_res_count = res_state[start_idx - 1].available_workers_count

                assert available_res_count >= res_count

                for event in res_state[start_idx: end_idx]:
                    assert event.available_workers_count >= res_count
                    event.available_workers_count -= res_count

                res_state.add(
                    ScheduleEvent(task_index, EventType.START, start_time, None, available_res_count - res_count)
                )

                end_idx = res_state.bisect_right(end_time) - 1
                # TODO Check correctness
                if res_state[end_idx].time == end_time:
                    end_count = res_state[end_idx].available_workers_count
                else:
                    end_count = res_state[end_idx].available_workers_count + res_count

                res_state.add(ScheduleEvent(task_index, EventType.END, end_time, None, end_count))

    def _validate(self, res_holder_id: str, res_info: tuple[str, int, Time, Time]):
        res_holder_state = self._timeline[res_holder_id]

        res_name, res_count, start_time, end_time = res_info
        res_state = res_holder_state[res_name]
        start_idx = res_state.bisect_right(start_time)

        end_idx = res_state.bisect_left((end_time, -1, EventType.INITIAL))
        available_res_count = res_state[start_idx - 1].available_workers_count

        assert available_res_count >= res_count

        for event in res_state[start_idx: end_idx]:
            assert event.available_workers_count >= res_count
            event.available_workers_count -= res_count



from _operator import itemgetter
from enum import Enum
from itertools import chain, groupby

from sampo.generator.environment.contractor import get_contractor
from sampo.schemas.contractor import Contractor
from sampo.schemas.graph import WorkGraph
from sampo.schemas.requirements import WorkerReq


class ContractorGenerationMethod(Enum):
    MIN = 'min'
    MAX = 'max'
    AVG = 'avg'


def _value_by_req(method: ContractorGenerationMethod, req: WorkerReq) -> int:
    """
    Sets the function by which the number for the function of searching for a contractor by the graph of works
    is determined by the given parameter

    :param method: type the specified parameter: min ~ min_count, max ~ max_count, avg ~ (min_count + max_count) / 2
    :param req: the Worker Req
    :return:
    """
    val = 0
    match method:
        case ContractorGenerationMethod.MIN: val = req.min_count
        case ContractorGenerationMethod.AVG: val = (req.min_count + req.max_count) / 2
        case ContractorGenerationMethod.MAX: val = req.max_count
    return int(val)


def get_contractor_by_wg(wg: WorkGraph,
                         scaler: float | None = 1,
                         method: ContractorGenerationMethod = ContractorGenerationMethod.AVG,
                         contractor_id: str | None = None,
                         contractor_name: str | None = None) -> Contractor:
    """
    Creates a pool of contractor resources based on job requirements, selecting the maximum specified parameter

    :param wg: The graph of works for which it is necessary to find a set of resources
    :param scaler: Multiplier for the number of resources in the contractor
    :param method: type the specified parameter: min ~ min_count, max ~ max_count, avg ~ (min_count + max_count) / 2
    :param contractor_id: generated contractor's id
    :param contractor_name: generated contractor's name
    :return: the contractor capable of completing given `WorkGraph`
    """
    if scaler < 1:
        raise ValueError('scaler should be greater than 1')
    wg_reqs = list(chain(*[n.work_unit.worker_reqs for n in wg.nodes]))
    min_wg_reqs = [(req.kind, _value_by_req(method, req)) for req in wg_reqs]
    maximal_min_req: dict[str, int] = \
        dict(list(group[1])[-1] for group in groupby(sorted(min_wg_reqs), itemgetter(0)))
    return get_contractor(scaler, sigma_scaler=0, worker_proportions=maximal_min_req,
                          contractor_id=contractor_id, contractor_name=contractor_name)



from collections import defaultdict
from typing import Iterable

from sampo.schemas import Worker, Contractor
from sampo.schemas.types import WorkerName, ContractorName

WorkerContractorPool = dict[WorkerName, dict[ContractorName, Worker]]


def get_worker_contractor_pool(contractors: Iterable[Contractor]) -> WorkerContractorPool:
    """
    Gets worker-contractor dictionary from contractors list.
    Alias for frequently used functionality.

    :param contractors: list of all the considered contractors
    :return: dictionary of workers by worker name, next by contractor id
    """
    worker_pool = defaultdict(dict)
    for contractor in contractors:
        for name, worker in contractor.workers.items():
            worker_pool[name][contractor.id] = worker.copy()
    return worker_pool



from collections import defaultdict
from dataclasses import dataclass, field
from typing import Union, Iterable
from uuid import uuid4

import numpy as np
from pandas import DataFrame, Series

from sampo.schemas.identifiable import Identifiable
from sampo.schemas.resources import Worker, Equipment
from sampo.schemas.serializable import AutoJSONSerializable
from sampo.schemas.types import WorkerName, ContractorName
from sampo.utilities.serializers import custom_serializer

DEFAULT_CONTRACTOR_CAPACITY = 25


@dataclass
class Contractor(AutoJSONSerializable['Contractor'], Identifiable):
    """
    Used to store information about the contractor and its resources
    :param workers: dictionary, where the key is the employee's specialty, and the value is the pool of employees of
    this specialty
    :param equipments: dictionary, where the key is the type of technique, and the value is the pool of techniques of
    that type
    """
    workers: dict[str, Worker] = field(default_factory=dict)
    equipments: dict[str, Equipment] = field(default_factory=dict)

    def __post_init__(self):
        for w in self.workers.values():
            w.contractor_id = self.id

    def __hash__(self) -> str:
        return hash(self.id)

    @custom_serializer('workers')
    def serialize_workers(self, value) -> list[dict]:
        return [{'key': k, 'val': v._serialize()} for k, v in value.items()]

    @custom_serializer('equipments')
    def serialize_equipment(self, value) -> []:
        return {k: v._serialize() for k, v in value.items()}

    @classmethod
    @custom_serializer('workers', deserializer=True)
    def deserialize_workers(cls, value) -> dict[str, Worker]:
        return {i['key']: Worker._deserialize(i['val']) for i in value}

    @classmethod
    @custom_serializer('equipments', deserializer=True)
    def deserialize_equipment(cls, value) -> dict[str,Equipment]:
        return {k: Equipment._deserialize(v) for k, v in value.items()}



from sampo.schemas.graph import GraphNode, WorkGraph
import numpy as np
import random


def map_lft_lst(wg: WorkGraph, node_id2duration: dict[str, int]) -> tuple[dict[str, int], dict[str, int]]:
    project_max_duration = sum(node_id2duration.values())
    nodes2lft = {wg.finish.id: project_max_duration}
    nodes2lst = {wg.finish.id: project_max_duration - node_id2duration[wg.finish.id]}

    for node in reversed(wg.nodes[:-1]):
        suc_lst = [nodes2lst[child.id] for child in node.children]
        nodes2lft[node.id] = min(suc_lst)
        nodes2lst[node.id] = nodes2lft[node.id] - node_id2duration[node.id]

    nodes = [node for node in wg.nodes if not node.is_inseparable_son()]

    nodes2lft = {node.id: nodes2lft[node.get_inseparable_chain_with_self()[-1].id] for node in nodes}
    nodes2lst = {node.id: nodes2lst[node.get_inseparable_chain_with_self()[0].id] for node in nodes}

    return nodes2lft, nodes2lst


def lft_prioritization(wg: WorkGraph, node_id2duration: dict[str, int]) -> list[GraphNode]:
    """
    Return nodes ordered by MIN-LFT priority rule.
    """
    nodes2lft, _ = map_lft_lst(wg, node_id2duration)

    ordered_nodes = sorted(nodes2lft.keys(), key=lambda node_id: nodes2lft[node_id], reverse=True)

    ordered_nodes = [wg[node_id] for node_id in ordered_nodes]

    return ordered_nodes


def lft_randomized_prioritization(wg: WorkGraph, node_id2duration: dict[str, int],
                                  rand: random.Random = random.Random()) -> list[GraphNode]:
    """
    Return ordered nodes sampled by MIN-LFT and MIN-LST priority rules.
    """

    def is_eligible(node_id):
        return parents[node_id].issubset(selected_ids_set)

    nodes2lft, nodes2lst = map_lft_lst(wg, node_id2duration)

    nodes = [wg[node_id] for node_id in nodes2lft.keys()]

    # map nodes and inseparable chain's head
    inseparable_parents = {}
    for node in nodes:
        for child in node.get_inseparable_chain_with_self():
            inseparable_parents[child.id] = node.id

    # map nodes and their children using only inseparable chain heads
    children = {node.id: set([inseparable_parents[child.id]
                              for inseparable in node.get_inseparable_chain_with_self()
                              for child in inseparable.children]) - {node.id}
                for node in nodes}

    # map nodes and their parents using only inseparable chain heads
    parents = {node.id: set() for node in nodes}
    for node, node_children in children.items():
        for child in node_children:
            parents[child].add(node)

    selected_ids = []
    selected_ids_set = set()
    candidates = {wg.start.id}

    while candidates:
        eligibles = [node_id for node_id in candidates if is_eligible(node_id)]

        priority_mapper = nodes2lft if rand.random() < 0.5 else nodes2lst

        weights = np.array([priority_mapper[node_id] for node_id in eligibles])
        weights = weights.max() - weights + 1
        weights = weights / weights.sum()

        selected_id = rand.choices(eligibles, weights=weights)[0]

        selected_ids.append(selected_id)
        selected_ids_set.add(selected_id)
        candidates.remove(selected_id)
        candidates.update([child_id for child_id in children[selected_id]])

    ordered_nodes = list(reversed([wg[node_id] for node_id in selected_ids]))

    return ordered_nodes
